#!/work/ROMO/anaconda3/bin/python
import sys
import numpy as np
from warnings import warn
from collections import defaultdict
from datetime import datetime, timedelta
from PseudoNetCDF.coordutil import getmap


def plot(ifiles, args):
    import matplotlib.pyplot as plt
    from PseudoNetCDF.coordutil import getpresbnds, getsigmabnds
    plt.rcParams['text.usetex'] = False
    from matplotlib.colors import LinearSegmentedColormap, BoundaryNorm, LogNorm, Normalize
    map = not args.nomap
    scale = args.scale
    minmax = eval(args.minmax)
    minmaxq = eval(args.minmaxq)
    sigma = args.sigma
    maskzeros = args.maskzeros
    try:
        f, = ifiles
    except Exception:
        raise ValueError(
            'curtain plot expects one file when done. Try stack time --stack=time to concatenate')

    if sigma:
        vertcrd = getsigmabnds(f)
    else:
        vertcrd = getpresbnds(f, pref=101325., ptop=getattr(f, 'VGTOP', 10000))
        if vertcrd.max() > 2000:
            vertcrd /= 100.

    reversevert = not (np.diff(vertcrd) > 0).all()
    for var_name in args.variables:
        temp = defaultdict(lambda: 1)
        try:
            eval(var_name, None, temp)
            var = eval(var_name, None, f.variables)[:]
        except Exception:
            temp[var_name]
            var = f.variables[var_name][:]
        if args.itertime:
            vars = [('time%02d' % vi, v) for vi, v in enumerate(var)]
        else:
            if var.shape[0] != 1:
                parser.print_help()
                sys.stderr.write(
                    '\n*****\n\nFile must have only one time or use the itertime options; to reduce file to one time, see the --slice and --reduce operators\n\n')
                exit()
            vars = [('', var[0])]

        for lstr, var in vars:
            bmap = None
            if maskzeros:
                var = np.ma.masked_values(var, 0)
            vmin, vmax = np.percentile(
                np.ma.compressed(var).ravel(), list(minmaxq))
            if minmax[0] is not None:
                vmin = minmax[0]
            if minmax[1] is not None:
                vmax = minmax[1]

            if args.normalize is not None:
                norm = eval(args.normalize)
                if norm.scaled():
                    vmin = norm.vmin
                    vmax = norm.vmax
            else:
                if scale == 'log':
                    bins = np.logspace(np.log10(vmin), np.log10(vmax), 11)
                elif scale == 'linear':
                    bins = np.linspace(vmin, vmax, 11)
                elif scale == 'deciles':
                    bins = np.percentile(np.ma.compressed(np.ma.masked_greater(np.ma.masked_less(var, vmin).view(
                        np.ma.MaskedArray), vmax)).ravel(), [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100])
                    bins[0] = vmin
                    bins[-1] = vmax
                norm = BoundaryNorm(bins, ncolors=256)

            if map:
                fig = plt.figure(figsize=(8, 8))
                fig.subplots_adjust(hspace=.3, wspace=.3)
                axmap = fig.add_subplot(3, 3, 5)
                try:
                    cmaqmap = getmap(f, resolution=args.resolution)
                    cmaqmap.drawcoastlines(ax=axmap)
                    cmaqmap.drawcountries(ax=axmap)
                    if not args.states:
                        cmaqmap.drawstates(ax=axmap)
                    if args.counties:
                        cmaqmap.drawcounties(ax=axmap)
                    cmaqmap.drawparallels(
                        np.arange(-90, 100, 10), labels=[True, True, False, False], ax=axmap)
                    cmaqmap.drawmeridians(
                        np.arange(-180, 190, 20), labels=[False, False, True, True], ax=axmap)
                except Exception as e:
                    warn('An error occurred and no map will be shown:\n%s' % str(e))
                axn = fig.add_subplot(3, 3, 2, sharex=axmap)
                axw = fig.add_subplot(3, 3, 4, sharey=axmap)
                axe = fig.add_subplot(3, 3, 6, sharey=axmap)
                axs = fig.add_subplot(3, 3, 8, sharex=axmap)
                cax = fig.add_axes([.8, .7, .05, .25])
                for ax in [axmap, axe]:
                    ax.yaxis.set_major_formatter(plt.NullFormatter())
                for ax in [axmap, axn]:
                    ax.xaxis.set_major_formatter(plt.NullFormatter())
                for ax in [axn, axs]:
                    if sigma:
                        ax.set_ylabel('sigma')
                    else:
                        ax.set_ylabel('pressure')
                for ax in [axe, axw]:
                    if sigma:
                        ax.set_xlabel('sigma')
                    else:
                        ax.set_xlabel('pressure')
                xyfactor = 1
            else:
                fig = plt.figure(figsize=(16, 4))
                fig.subplots_adjust(bottom=0.15)
                axw = fig.add_subplot(1, 4, 1)
                axn = fig.add_subplot(1, 4, 2)
                axe = fig.add_subplot(1, 4, 3)
                axs = fig.add_subplot(1, 4, 4)
                cax = fig.add_axes([.91, .1, .025, .8])
                if sigma:
                    axw.set_ylabel('sigma')
                else:
                    axw.set_ylabel('pressure')

                xyfactor = 1e-3  # m -> km

            x = f.NCOLS + 1
            y = f.NROWS + 1
            start_south = 0
            end_south = start_south + x
            start_east = end_south
            end_east = start_east + y
            start_north = end_east
            end_north = start_north + x
            start_west = end_north
            end_west = start_west + y
            X, Y = np.meshgrid(np.arange(x + 1) * f.XCELL * xyfactor, vertcrd)
            # patchess =
            axs.pcolor(
                X, Y, var[:, start_south:end_south], cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
            if not map:
                if reversevert:
                    axs.set_ylim(*axs.get_ylim()[::-1])
                axs.set_title('South')
                axs.set_xlabel('E to W km')
                axs.set_xlim(*axs.get_xlim()[::-1])
            else:
                if not reversevert:
                    axs.set_ylim(*axs.get_ylim()[::-1])

            X, Y = np.meshgrid(np.arange(-1, x) * f.XCELL * xyfactor, vertcrd)
            # patchesn =
            axn.pcolor(
                X, Y, var[:, start_north:end_north], cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
            if reversevert:
                axn.set_ylim(*axn.get_ylim()[::-1])
            if not map:
                axn.set_title('North')
                axn.set_xlabel('W to E km')

            if map:
                X, Y = np.meshgrid(vertcrd, np.arange(y + 1) * f.YCELL)
                # patchese =
                axe.pcolor(X, Y, var[:, start_east:end_east].swapaxes(
                    0, 1), cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
                if reversevert:
                    axe.set_xlim(*axe.get_xlim()[::-1])
            else:
                X, Y = np.meshgrid(np.arange(y + 1) *
                                   f.YCELL * xyfactor, vertcrd)
                # patchese =
                axe.pcolor(
                    X, Y, var[:, start_east:end_east], cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
                if reversevert:
                    axe.set_ylim(*axe.get_ylim()[::-1])
                axe.set_title('East')
                axe.set_xlabel('N to S km')
                axe.set_xlim(*axe.get_xlim()[::-1])
            if map:
                X, Y = np.meshgrid(vertcrd, np.arange(-1, y) * f.YCELL)
                patchesw = axw.pcolor(X, Y, var[:, start_west:end_west].swapaxes(
                    0, 1), cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
                if not reversevert:
                    axw.set_xlim(*axw.get_xlim()[::-1])
            else:
                X, Y = np.meshgrid(np.arange(-1, y) *
                                   f.YCELL * xyfactor, vertcrd)
                patchesw = axw.pcolor(
                    X, Y, var[:, start_west:end_west], cmap=bmap, vmin=vmin, vmax=vmax, norm=norm)
                if reversevert:
                    axw.set_ylim(*axw.get_ylim()[::-1])
                axw.set_title('West')
                axw.set_xlabel('S to N km')
            if map:
                for ax in [axe, axw]:
                    ax.axis('tight', axis='x')
                    plt.setp(ax.xaxis.get_majorticklabels(), rotation=90)
                for ax in [axs, axn]:
                    ax.axis('tight', axis='y')
            else:
                for ax in [axe, axn, axw, axs] + ([axmap] if map else []):
                    ax.axis('tight')

            if 'TFLAG' in f.variables.keys():
                SDATE = f.variables['TFLAG'][:][0, 0, 0]
                EDATE = f.variables['TFLAG'][:][-1, 0, 0]
                STIME = f.variables['TFLAG'][:][0, 0, 1]
                ETIME = f.variables['TFLAG'][:][-1, 0, 1]
                if SDATE == 0:
                    SDATE = 1900001
                    EDATE = 1900001
                try:
                    sdate = datetime.strptime(
                        '%07d %06d' % (SDATE, STIME), '%Y%j %H%M%S')
                    edate = datetime.strptime(
                        '%07d %06d' % (EDATE, ETIME), '%Y%j %H%M%S')
                except Exception as e:
                    warn('Unable to convert time:' + str(e))
            elif 'tau0' in f.variables.keys():
                sdate = datetime(1985, 1, 1, 0) + \
                    timedelta(hours=f.variables['tau0'][0])
                edate = datetime(1985, 1, 1, 0) + \
                    timedelta(hours=f.variables['tau1'][-1])
            else:
                sdate = datetime(1900, 1, 1, 0)
                edate = datetime(1900, 1, 1, 0)
            try:
                title = '%s %s to %s' % (var_name, sdate.strftime(
                    '%Y-%m-%d'), edate.strftime('%Y-%m-%d'))
            except Exception:
                title = var_name
            fig.suptitle(title.replace('O3', 'Ozone at Regional Boundaries'))
            if var.min() < vmin and var.max() > vmax:
                extend = 'both'
            elif var.min() < vmin:
                extend = 'min'
            elif var.max() > vmax:
                extend = 'max'
            else:
                extend = 'neither'
            fig.colorbar(patchesw, cax=cax, extend=extend)
            cax.set_xlabel('ppm')
            figpath = args.outpath + var_name + lstr + '.' + args.figformat
            fig.savefig(figpath)
            if args.verbose > 0:
                print('Saved fig', figpath)
            plt.close(fig)


if __name__ == '__main__':
    from PseudoNetCDF.pncparse import getparser, pncparse

    parser = getparser(has_ofile=True, map_options=True,
                       plot_options=True, interactive=False)
    parser.add_argument("--itertime", dest="itertime", action="store_true", default=False,
                        help="Create plots for each time.")

    parser.add_argument("-n", "--no-map", dest="nomap", action="store_true", default=False,
                        help="Try to plot with map")

    parser.add_argument("--sigma", dest="sigma", action="store_true", default=False,
                        help="Plot data on sigma coordinate.")

    parser.add_argument("--scale", dest="scale", type=str, default='deciles',
                        help="Defaults to deciles (i.e., 10 equal probability bins), but linear and log are also options.")

    parser.add_argument("--mask-zeros", dest="maskzeros", action="store_true", default=False,
                        help="Defaults False.")

    parser.add_argument("--minmax", dest="minmax", type=str, default="None,None",
                        help="Defaults None, None.")

    parser.add_argument("--minmaxq", dest="minmaxq", type=str, default='0,100',
                        help="Defaults 0,100.")
    parser.epilog = """
Example:
    $ pncboundaries.py outputs/ts20120301.bpch.BCON.nc -v O3 test_boundary -r TSTEP,mean --states
"""
    ifiles, args = pncparse(has_ofile=True, parser=parser)
    if args.variables is None:
        raise ValueError('User must specify variable(s) to plot:\n%s' %
                         '\n\t'.join(ifiles[0].variables.keys()))

    plot(ifiles, args)



__all__ = ['ncf2height_pressure', 'write_hgtprss']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx height_pressure writer
============================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx
              height_pressure files.
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Site-Packages
from numpy import array

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import writeline
from PseudoNetCDF._getwriter import registerwriter


def ncf2height_pressure(ncffile, outpath, hght='HGHT', pres='PRES',
                        tflag='TFLAG'):
    outfile = open(outpath, 'wb')
    thp = zip(ncffile.variables[tflag][:, 0, :],
              ncffile.variables[hght],
              ncffile.variables[pres])
    for (d, t), h3d, p3d in thp:
        t = array(t.astype('>f') / 100, ndmin=1).astype('>f')
        d = array(d, ndmin=1).astype('>i')
        d = (d % (d // 100000 * 100000)).astype('>i')
        for i, (h2d, p2d) in enumerate(zip(h3d, p3d)):
            h2d = h2d.astype('>f')
            p2d = p2d.astype('>f')
            buf = array((h2d.size + 2) * 4, ndmin=1).astype('>i').tostring()
            outfile.write(buf + t.tostring() + d.tostring())
            h2d.tofile(outfile)
            outfile.write(buf)
            outfile.write(buf + t.tostring() + d.tostring())
            p2d.tofile(outfile)
            outfile.write(buf)
    outfile.flush()
    return outfile


registerwriter('camxfiles.height_pressure', ncf2height_pressure)
registerwriter('height_pressure', ncf2height_pressure)


def write_hgtprss(sdate, stime, time_step, vals):
    """
    Takes an iterable and some defining information
    and creates a CAMx read wind file

    sdate - integer start date
    stime - float start time
    time_step - float increment between times

    vals - array axes time,uv,xy,z
    """
    hp_string = ""
    edate, etime = timeadd((sdate, stime), (0, vals.shape[0] * time_step))

    tind = (2400, 24)[int(time_step % 2)]
    tenum = enumerate(timerange((sdate, stime),
                                timeadd((edate, etime), (0, 0)),
                                time_step,
                                tind))
    for i, (d, t) in tenum:
        for k in range(vals.shape[-1]):
            for hp in range(2):
                hpvals = [t, d]
                hpvals.extend(vals[i, hp, ..., k])
                hp_string += writeline(hpvals, "fi" + "f" * vals.shape[-2])
    return hp_string



from __future__ import print_function
from PseudoNetCDF import warn
import numpy as np
from collections import OrderedDict


def getbounds(ifile, dimkey):
    dim = ifile.dimensions[dimkey]
    dimboundkey = dimkey + '_bounds'
    dimbndkey = dimkey + '_bnds'
    if dimboundkey in ifile.variables:
        db = ifile.variables[dimboundkey]
    elif dimbndkey in ifile.variables:
        db = ifile.variables[dimbndkey]
    elif dimkey in ifile.variables:
        d = ifile.variables[dimkey]
        dd = np.diff(d)
        ddm = dd.mean()
        if not (ddm == dd).all():
            warn('Bounds is an approximation assuming ' +
                 '%s variable is cell centers' % dimkey)

        db = (d[:-1] + d[1:]) / 2.
        db = np.append(np.append(d[0] - dd[0] / 2., db), d[-1] + dd[-1] / 2)
        return db
    else:
        return np.arange(0, len(dim) + 1)

    if len(dim) == db.shape[0] and db.shape[1] == 2:
        return np.append(db[:, 0], db[-1, 1])
    elif db.ndim == 1:
        return db
    else:
        return db[:, 0]


def getlonlatcoordstr(ifile, makemesh=None):
    """
    ifile - file with latitude and longitude variables
    makemesh - use numpy.meshgrid to construct gridded values (default None)
               None - check if longitude and latitude are coordinate variables
                      or have different dimensions if so set to True
               True - use meshgrid
               False - assume latitude and longitude are on same g
    """
    lon = ifile.variables['longitude']
    lat = ifile.variables['latitude']

    if (
        lon.dimensions != lat.dimensions or
        (lon.dimensions == ('longitude',) and lat.dimensions == ('latitude',))
    ):
        lon, lat = np.meshgrid(lon[:], lat[:])

    return '/'.join(['%s,%s' % ll for ll in zip(lon[:].flat, lat[:].flat)])


def _parse_ref_date(base):
    from datetime import datetime
    fmts = [
        # has time and Z
        (lambda x: x.count(':') == 2 and x[-3:] ==
         'UTC', '+0000', '%Y-%m-%d %H:%M:%S UTC%z'),
        # has time and Z
        (lambda x: x.count(':') == 1 and x[-3:] ==
         'UTC', '+0000', '%Y-%m-%d %H:%M UTC%z'),
        (lambda x: 'UTC' in x, '+0000', '%Y-%m-%d %H UTC%z'),  # has hour and Z
        # has time and Z
        (lambda x: x.count(':') ==
         2 and x[-1] == 'Z', '+0000', '%Y-%m-%d %H:%M:%SZ%z'),
        # has time and Z
        (lambda x: x.count(':') ==
         1 and x[-1] == 'Z', '+0000', '%Y-%m-%d %H:%MZ%z'),
        (lambda x: 'Z' == x[-1], '+0000',
         '%Y-%m-%d %HZ%z'),  # has hour and Z
        # full ISO8601 datetime with numeric timezone info
        (None, '', '%Y-%m-%d %H:%M:%S%z'),
        (None, '+0000', '%Y-%m-%d %H:%M:%S%z'),  # missing timezone
        # missing seconds and timezone
        (None, ':00+0000', '%Y-%m-%d %H:%M:%S%z'),
        (None, '', '%Y-%m-%d %H:%M%z'),  # missing seconds
        # missing minutes and seconds and timezone
        (None, ':00:00+0000', '%Y-%m-%d %H:%M:%S%z'),
        (None, '', '%Y-%m-%d %H%z'),  # missing minutes and seconds
        # missing time and timezone
        (None, ' 00:00:00+0000', '%Y-%m-%d %H:%M:%S%z'),
        (None, '', '%Y-%m-%d %z'),  # missing time
        (None, '', '%Y-%m-%d%z'),  # missing time
    ]
    for opti, (test, suffix, fmt) in enumerate(fmts):
        if test is None or test(base):
            try:
                rdate = datetime.strptime(base + suffix, fmt)
                return rdate
            except Exception:
                pass
    else:
        # try using netcdftime
        try:
            import netcdftime
            ut = netcdftime.netcdftime.utime(base.strip())
            sdate = ut.num2date(0.)
            return sdate
        except Exception as e:
            raise ValueError('Could not find appropriate date; tried and ' +
                             'failed to use netcdftime' + str(e))


def gettimes(ifile):
    """
    Converts relative time to datetime objects
     - Finds time variable (e.g., time or TFLAG, tau0)
     - Parses reference date
     - Converts
    """
    from datetime import datetime, timedelta
    if 'time' in ifile.variables.keys():
        time = ifile.variables['time']
        if 'since' in time.units:
            unit, base = time.units.strip().split(' since ')
            sdate = _parse_ref_date(base)
            out = sdate + \
                np.array([timedelta(**{unit: float(i)}) for i in time[:]])
            return out
        else:
            return time
    elif 'TFLAG' in ifile.variables.keys():
        dates = ifile.variables['TFLAG'][:][:, 0, 0]
        times = ifile.variables['TFLAG'][:][:, 0, 1]
        yyyys = (dates // 1000).astype('i')
        jjj = dates % 1000
        hours = times // 10000
        minutes = times % 10000 // 100
        seconds = times % 100
        days = jjj + (hours + minutes / 60. + seconds / 3600.) / 24.
        out = np.array([datetime(yyyy, 1, 1) + timedelta(days=day - 1)
                        for yyyy, day in zip(yyyys, days)])
        return out
    elif 'tau0' in ifile.variables.keys():
        out = (datetime(1985, 1, 1, 0) +
               np.array([timedelta(hours=i)
                         for i in ifile.variables['tau0'][:]]))
        return out
    else:
        raise ValueError('cannot understand time for file')


def gettimebnds(ifile):
    from datetime import datetime, timedelta
    if 'TFLAG' in ifile.variables.keys():
        dates = ifile.variables['TFLAG'][:][:, 0, 0]
        times = ifile.variables['TFLAG'][:][:, 0, 0]
        yyyys = (dates // 1000).astype('i')
        jjj = dates % 1000
        hours = times // 10000
        minutes = times % 10000 // 100
        seconds = times % 100
        days = jjj + (hours + minutes / 60. + seconds / 3600.) / 24.
        out = np.array([datetime(yyyy, 1, 1) + timedelta(days=day - 1)
                        for yyyy, day in zip(yyyys, days)])

        hours = ifile.TSTEP // 10000
        minutes = ifile.TSTEP % 10000 // 100
        seconds = ifile.TSTEP % 100
        hours = (hours + minutes / 60. + seconds / 3600.)
        return np.array([out, out + timedelta(hours=hours)]).T
    elif 'tau0' in ifile.variables.keys() and 'tau1' in ifile.variables.keys():
        out1 = (datetime(1985, 1, 1, 0) +
                np.array([timedelta(hours=i)
                         for i in ifile.variables['tau0'][:]]))
        out2 = (datetime(1985, 1, 1, 0) +
                np.array([timedelta(hours=i)
                          for i in ifile.variables['tau1'][:]]))
        return np.array([out1, out2]).T
    elif 'time' in ifile.variables.keys():
        time = ifile.variables['time']
        if 'since' in time.units:
            unit, base = time.units.strip().split(' since ')
            sdate = _parse_ref_date(base)
            out = sdate + \
                np.array([timedelta(**{unit: float(i)}) for i in time[:]])
            if len(out) > 1:
                dt = (out[1] - out[0])
            else:
                dt = timedelta(**{unit: 0.})

            return np.array([out, out + dt]).T
        else:
            return np.array([time, time + (time[1] - time[0])]).T
    else:
        raise ValueError('cannot understand time for file')


def getsigmabnds(ifile):
    if hasattr(ifile, 'VGLVLS'):
        return ifile.VGLVLS[:]
    elif 'etai_pressure' in ifile.variables:
        etai_pressure = ifile.variables['etai_pressure']
        return ((etai_pressure - etai_pressure[-1]) /
                (etai_pressure[0] - etai_pressure[-1]))
    elif 'layer_bounds' in ifile.variables:
        lay = ifile.variables['layer_bounds']
        if lay.units.strip() in ('Pa', 'hPa'):
            sigma = (lay[:] - lay[-1]) / (lay[0] - lay[-1])
            return sigma
        else:
            warn("Unknown tranform of layer to sigma; " +
                 "sigma units %s" % lay.units)
            return lay
    else:
        warn("Unknown vertical coordinate")
        if hasattr(ifile, 'NLAYS'):
            nlays = ifile.NLAYS
        elif 'LAY' in ifile.dimensions:
            nlays = len(ifile.dimensions['LAY'])
        elif 'lev' in ifile.dimensions:
            nlays = len(ifile.dimensions['lev'])
        elif 'layer' in ifile.dimensions:
            nlays = len(ifile.dimensions['layer'])
        else:
            nlays = 1
        return np.arange(nlays)


def pres_from_sigma(sigma, pref, ptop, avg=False):
    pres = sigma * (pref - ptop) + ptop
    if avg:
        pres = pres[:-1] + np.diff(pres) / 2.
    return pres


def getpresmid(ifile, pref=101325., ptop=None):
    presb = getpresbnds(ifile, pref=101325., ptop=None)
    return presb[:-1] + np.diff(presb) / 2


def getsigmamid(ifile):
    sigmab = getsigmabnds(ifile)
    return sigmab[:-1] + np.diff(sigmab) / 2


def getpresbnds(ifile, pref=101325., ptop=None):
    if 'etai_pressure' in ifile.variables:
        return ifile.variables['etai_pressure'][:]
    elif 'layer_bounds' in ifile.variables:
        return ifile.variables['layer_bounds'][:]
    else:
        sigma = getsigmabnds(ifile)
        if ptop is None:
            if hasattr(ifile, 'VGTOP'):
                ptop = ifile.VGTOP
            else:
                warn("Assuming VGTOP = 10000 Pa")
                ptop = 10000

        return pres_from_sigma(sigma, pref=pref, ptop=ptop)


def getlatbnds(ifile):
    if 'latitude_bounds' in ifile.variables:
        latb = ifile.variables['latitude_bounds']
        unit = latb.units.strip()
        if 'nv' in latb.dimensions:
            if latb[:].ndim == 2 and len(ifile.dimensions['nv']) == 2:
                latb = np.append(latb[:][:, 0], latb[:][-1, 1])
            elif latb[:].ndim == 2 and len(ifile.dimensions['nv']) == 4:
                latb = np.append(latb[:][:, 0], latb[:][-1, 1])
            elif latb.ndim == 3:
                latb = latb[:, :, 0]

    elif 'latitude' in ifile.variables:
        latb = ifile.variables['latitude']
        unit = latb.units.strip()
        latb = latb[:]
        latdiff = np.diff(latb, axis=0)
        if not (latdiff == latdiff[[0]]).all():
            warn('Latitude bounds are approximate')
        latb = np.apply_along_axis(np.convolve, 0, latb, [0.5, 0.5])
        latb[0] *= 2
        latb[-1] *= 2
        if latb.ndim == 2:
            latb = np.append(latb, latb[:, [-1]], axis=1)

    elif 'ROW' in ifile.dimensions:
        unit = 'x (m)'
        latb = (np.arange(len(ifile.dimensions['ROW']) + 1) *
                getattr(ifile, 'YCELL', 1) / 1000.)
    else:
        raise KeyError('latitude bounds not found')
    return latb, unit


def getybnds(ifile):
    if 'ROW' in ifile.dimensions:
        unit = 'y (m)'
        latb = np.arange(
            len(ifile.dimensions['ROW']) + 1) * getattr(ifile, 'YCELL', 1)
    elif 'south_north' in ifile.dimensions:
        unit = 'y (m)'
        latb = np.arange(
            len(ifile.dimensions['south_north']) + 1) * getattr(ifile, 'DY', 1)
    else:
        raise KeyError('latitude bounds not found')
    return latb, unit


def getlonbnds(ifile):
    if 'longitude_bounds' in ifile.variables:
        lonb = ifile.variables['longitude_bounds']
        unit = lonb.units.strip()
        if 'nv' in lonb.dimensions:
            if lonb[:].ndim == 2 and len(ifile.dimensions['nv']) == 2:
                lonb = np.append(lonb[:][:, 0], lonb[:][-1, 1])
            elif lonb[:].ndim == 3:
                lonb = lonb[:][:, :, 0]
    elif 'longitude' in ifile.variables:
        lonb = ifile.variables['longitude']
        unit = lonb.units.strip()
        lonb = lonb[:]
        if lonb.ndim > 1:
            londiff = np.diff(lonb, axis=1)
            alldiffsame = (londiff == londiff[:, [0]]).all()
        elif lonb.ndim == 1:
            alldiffsame = True
            londiff = np.diff(lonb)
        else:
            raise ValueError(
                "Cannot infer longitude bounds when dimensions >2")
        if not alldiffsame:
            londiff = np.diff(lonb, axis=1)
            if not (londiff == londiff[:, [0]]).all():
                warn('Longitude bounds are approximate')
            lonb = (np.concatenate([lonb, lonb[:, [-1]]], axis=1) -
                    .5 * np.concatenate([londiff[:, :],
                                         londiff[:, [-1]],
                                         -londiff[:, [-1]]], axis=1))
            lonb = np.append(lonb, lonb[[-1], :], axis=0)
        else:
            londiff = np.diff(lonb, axis=0)
            lonb = (np.concatenate([lonb, lonb[[-1]]], axis=0) -
                    .5 * np.concatenate([londiff[:],
                                         londiff[[-1]],
                                         -londiff[[-1]]], axis=0))

    else:
        raise KeyError('longitude bounds not found')
    return lonb, unit


def getxbnds(ifile):
    if 'COL' in ifile.dimensions:
        unit = 'x (m)'
        lonb = np.arange(
            len(ifile.dimensions['COL']) + 1) * getattr(ifile, 'XCELL', 1)
    elif 'west_east' in ifile.dimensions:
        unit = 'x (m)'
        lonb = np.arange(
            len(ifile.dimensions['west_east']) + 1) * getattr(ifile, 'DX', 1)
    else:
        raise KeyError('x bounds not found')
    return lonb, unit


def getcdo(ifile):
    """
    ifile - file containing latitude, longitude and optionally latitude_bounds
            and longitude_bounds
    """
    import textwrap

    def wrapper(first, instr):
        outstr = "\n".join(textwrap.wrap(
            instr, width=72, subsequent_indent=' ' * 12, initial_indent=first))
        return outstr

    outdict = {}
    if 'latitude' in ifile.dimensions and 'longitude' in ifile.dimensions:
        outdict['gridtype'] = 'lonlat'
        outdict['nverts'] = 2
        outdict['NCOLS'] = len(ifile.dimensions['longitude'])
        outdict['NROWS'] = len(ifile.dimensions['latitude'])
    elif 'ROW' in ifile.dimensions and 'COL' in ifile.dimensions:
        outdict['gridtype'] = 'curvilinear'
        outdict['nverts'] = 4
        outdict['NCOLS'] = len(ifile.dimensions['COL'])
        outdict['NROWS'] = len(ifile.dimensions['ROW'])
    elif 'south_north' in ifile.dimensions and 'west_east' in ifile.dimensions:
        outdict['gridtype'] = 'curvilinear'
        outdict['nverts'] = 4
        outdict['NCOLS'] = len(ifile.dimensions['west_east'])
        outdict['NROWS'] = len(ifile.dimensions['south_north'])
    else:
        raise ValueError('Could not find latitude/longitude or ROW/COL')
    outdict['NCELLS'] = outdict['NCOLS'] * outdict['NROWS']
    LONSTR = ' '.join(' '.join(['%f' % lon for lon in row])
                      for row in ifile.variables['longitude'][:])
    LATSTR = ' '.join(' '.join(['%f' % lat for lat in row])
                      for row in ifile.variables['latitude'][:])
    LONBSTR = ' '.join(
        ' '.join([' '.join(['%f' % lon for lon in cell])
                  for cell in row])
        for row in ifile.variables['longitude_bounds'][:])
    LATBSTR = ' '.join(
        ' '.join([' '.join(['%f' % lat for lat in cell])
                  for cell in row])
        for row in ifile.variables['latitude_bounds'][:])
    outdict['LONSTR'] = wrapper('xvals     = ', LONSTR)
    outdict['LONBSTR'] = wrapper('xbounds   = ', LONBSTR)
    outdict['LATSTR'] = wrapper('yvals     = ', LATSTR)
    outdict['LATBSTR'] = wrapper('ybounds   = ', LATBSTR)
    return """
    gridtype  = curvilinear
    nvertex   = %(nverts)d
    gridsize  = %(NCELLS)d
    xsize     = %(NCOLS)d
    ysize     = %(NROWS)d
    xunits    = degrees_east
    yunits    = degrees_north
    %(LONSTR)s
    %(LONBSTR)s
    %(LATSTR)s
    %(LATBSTR)s
    """ % outdict


def getprojwkt(ifile, withgrid=False):
    import osr
    proj4str = getproj4(ifile, withgrid=withgrid)

    srs = osr.SpatialReference()
    # Imports WKT to Spatial Reference Object
    srs.ImportFromProj4(proj4str)
    srs.ExportToWkt()  # converts the WKT to an ESRI-compatible format
    return srs.ExportToWkt()


def basemap_from_file(ifile, withgrid=False, **kwds):
    """
    Typically, the user will need to provide some options
    """
    proj4 = getproj4(ifile, withgrid=withgrid)
    basemap_options = basemap_options_from_proj4(proj4, **kwds)
    if 'llcrnrx' in basemap_options:
        if 'urcrnrx' in kwds:
            basemap_options['urcrnrx'] = kwds['urcrnrx']
        elif 'width' in kwds:
            basemap_options['urcrnrx'] = basemap_options['llcrnrx'] + \
                kwds['width']
        elif 'x' in ifile.variables:
            x = ifile.variables['x']
            urx = x.max() + np.mean(np.diff(x))
            basemap_options['urcrnrx'] = urx
        else:
            raise KeyError(
                'When a false_easting is available, the file must contain ' +
                'an x variable or the user must supply width or urcrnrx')
    if 'llcrnry' in basemap_options:
        if 'urcrnry' in kwds:
            basemap_options['urcrnry'] = kwds['urcrnry']
        elif 'height' in kwds:
            basemap_options['urcrnry'] = basemap_options['llcrnry'] + \
                kwds['height']
        elif 'y' in ifile.variables:
            y = ifile.variables['y']
            ury = y.max() + np.mean(np.diff(y))
            basemap_options['urcrnry'] = ury
        else:
            raise KeyError(
                'When a false_northing is available, the file must contain ' +
                'a y variable or the user must supply height or urcrnry')

    from mpl_toolkits.basemap import Basemap
    print(basemap_options)
    bmap = Basemap(**basemap_options)
    return bmap


def basemap_options_from_proj4(proj4, **kwds):
    """
    proj4 - string with projection optoins according to the proj4 system
    kwds - add keywords to control basemap specific options
        resolution = 'i' or 'c' or 'h' controls dpi of boundaries
        llcrnrlon=None, llcrnrlat=None, urcrnrlon=None, urcrnrlat=None,
        llcrnrx=None, llcrnry=None, urcrnrx=None, urcrnry=None,
        width=None, height=None,
    """
    excluded = ('proj', 'a', 'b', 'x_0', 'y_0', 'to_meter')
    proj4_options = OrderedDict()
    for seg in proj4.split():
        if '=' in seg:
            k, v = seg.split('=')
            if k in ('+proj', '+ellps'):
                v = '"' + v + '"'
            proj4_options[k.replace('+', '')] = eval(v)

    basemap_options = dict(
        [(k, v) for k, v in proj4_options.items() if k not in excluded])
    basemap_options['projection'] = proj4_options['proj']
    if 'a' in proj4_options and 'b' in proj4_options:
        basemap_options['rsphere'] = (proj4_options['a'], proj4_options['b'])
    elif 'a' in proj4_options and 'f' in proj4_options:
        basemap_options['rsphere'] = (
            proj4_options['a'],
            -(proj4_options['f'] * proj4_options['a'] - proj4_options['a']))
    elif 'a' in proj4_options:
        basemap_options['rsphere'] = (proj4_options['a'], proj4_options['a'])

    if 'x_0' in proj4_options:
        basemap_options['llcrnrx'] = -proj4_options['x_0']
    if 'y_0' in proj4_options:
        basemap_options['llcrnry'] = -proj4_options['y_0']
    basemap_options.update(**kwds)
    return basemap_options


def basemap_from_proj4(proj4, **kwds):
    from mpl_toolkits.basemap import Basemap
    basemap_options = basemap_options_from_proj4(proj4, **kwds)
    if basemap_options['projection'] in ('lonlat', 'longlat'):
        basemap_options['projection'] = 'cyl'
    bmap = Basemap(**basemap_options)
    return bmap


def getproj4_from_cf_var(gridmapping, withgrid=False):
    mapstr_bits = OrderedDict()
    gname = getattr(gridmapping, 'grid_mapping_name')
    pv4s = dict(lambert_conformal_conic='lcc',
                rotated_latitude_longitude='ob_tran',
                latitude_longitude='eqc',
                transverse_mercator='tmerc',
                equatorial_mercator='merc',
                mercator='merc',
                polar_stereographic='stere'
                )
    pv4name = pv4s[gname]
    for pk in gridmapping.ncattrs():
        pv = getattr(gridmapping, pk)
        if pk == 'grid_mapping_name':
            pv4 = pv4s[pv]
            mapstr_bits['proj'] = pv4
            if pv == 'rotated_latitude_longitude':
                mapstr_bits['o_proj'] = 'eqc'
        elif pk == 'standard_parallel':
            if pv4name == 'stere':
                if np.isscalar(pv):
                    mapstr_bits['lat_ts'] = float(pv)
                else:
                    mapstr_bits['lat_ts'] = pv[0]
            else:
                mapstr_bits['lat_1'] = pv[0]
            if not np.isscalar(pv) and len(pv) > 1:
                mapstr_bits['lat_2'] = pv[1]
        elif pk in (
            'longitude_of_central_meridian',
            'straight_vertical_longitude_from_pole'
        ):
            mapstr_bits['lon_0'] = pv
        elif pk == 'latitude_of_projection_origin':
            mapstr_bits['lat_0'] = pv
        elif pk == 'false_easting':
            mapstr_bits['x_0'] = pv
        elif pk == 'false_northing':
            mapstr_bits['y_0'] = pv
        elif pk == 'scale_factor_at_projection_origin':
            mapstr_bits['k_0'] = pv
        elif pk == 'earth_radius':
            mapstr_bits['a'] = pv
            mapstr_bits['b'] = pv
        elif pk == 'semi_major_axis':
            mapstr_bits['a'] = pv
        elif pk == 'semi_minor_axis':
            mapstr_bits['b'] = pv
        elif pk == 'inverse_flattening':
            mapstr_bits['f'] = 1 / pv
        elif pk == 'grid_north_pole_latitude':
            mapstr_bits['o_lat_p'] = pv
        elif pk == 'grid_north_pole_longitude':
            mapstr_bits['lon_0'] = pv
        else:
            warn('Currently not using:' + str(pk) + ' ' + str(pv))

    # repr is required to prevent rounding of numpy array values
    mapstr = ' '.join(['+%s=%s' % (k, v if isinstance(v, str) else repr(v))
                       for k, v in mapstr_bits.items()])
    return mapstr


def getproj(ifile, withgrid=False):
    import pyproj
    proj4str = getproj4(ifile, withgrid=withgrid)
    preserve_units = withgrid
    # pyproj adds +units=m, which is not right for latlon/lonlat
    if '+proj=lonlat' in proj4str or '+proj=latlon' in proj4str:
        preserve_units = True
    return pyproj.Proj(proj4str, preserve_units=preserve_units)


def getproj4(ifile, withgrid=False):
    """
    Arguments:
      ifile - PseudoNetCDF file
      withgrid - True to include gridding parameters

    Returns:
      proj4str - string with proj4 parameters
    """
    from .conventions.ioapi import getmapdef
    if (
        getattr(ifile, 'GDTYP', 0) in (1, 2, 6, 7) and
        all([hasattr(ifile, k)
             for k in 'P_GAM P_ALP P_BET XORIG YORIG XCELL YCELL'.split()])
    ):
        gridmapping = getmapdef(ifile, add=False)
        mapstr = getproj4_from_cf_var(gridmapping, withgrid=withgrid)
        if withgrid:
            dx = min(ifile.XCELL, ifile.YCELL)
            if ifile.XCELL != ifile.YCELL:
                warn('Grid is not regular: using minimum {}'.format(dx))
            if gridmapping.grid_mapping_name == 'latitude_longitude':
                er = min(
                    gridmapping.semi_minor_axis,
                    gridmapping.semi_major_axis
                )
                if (
                    gridmapping.semi_minor_axis !=
                    gridmapping.semi_major_axis
                ):
                    warn(
                        'Earth not a perfect sphere: using minimum ' +
                        '{}'.format(er)
                    )
                mapstr += ' +to_meter=%s' % (np.radians(1) * er * dx)
            else:
                mapstr += ' +to_meter=%s' % ifile.XCELL
    elif (
        getattr(ifile, 'Conventions',
                getattr(ifile, 'CONVENTIONS', ''))[:2].upper() == 'CF'
    ):
        gridmappings = []
        for k, v in ifile.variables.items():
            if hasattr(v, 'grid_mapping'):
                gridmappings.append(getattr(v, 'grid_mapping'))

        if len(gridmappings) == 0:
            warn('No known grid mapping; assuming lonlat')
            mapstr = '+proj=lonlat'
        else:
            gridmappings = list(set(gridmappings))
            if len(gridmappings) > 1:
                warn('Using first grid mapping of ' + str(gridmappings))
            if not gridmappings[0] in ifile.variables:
                warn(gridmappings[0] + ' could not be found; assuming lonlat')
                mapstr = '+proj=lonlat'
            else:
                gridmapping = ifile.variables[gridmappings[0]]
                mapstr = getproj4_from_cf_var(gridmapping, withgrid=withgrid)
    else:
        warn('No known grid mapping; assuming lonlat')
        mapstr = '+proj=lonlat'
    mapstr += ' +no_defs'
    return mapstr


def getmap(ifile, resolution='i'):
    from mpl_toolkits.basemap import Basemap
    from .conventions.ioapi import get_ioapi_sphere
    if (
        getattr(ifile, 'GDTYP', 0) in (2, 6, 7) and
        all([hasattr(ifile, k)
             for k in 'P_GAM P_ALP P_BET XORIG YORIG XCELL YCELL'.split()])
    ):
        try:
            NROWS = len(ifile.dimensions['ROW'])
            NCOLS = len(ifile.dimensions['COL'])
        except KeyError:
            NROWS = ifile.NROWS
            NCOLS = ifile.NCOLS

        llcrnrx = ifile.XORIG
        urcrnrx = ifile.XORIG + NCOLS * ifile.XCELL

        llcrnry = ifile.YORIG
        urcrnry = ifile.YORIG + NROWS * ifile.YCELL
        semi_major_axis, semi_minor_axis = get_ioapi_sphere()
        if ifile.GDTYP == 2:
            from mpl_toolkits.basemap import pyproj
            p = pyproj.Proj(proj='lcc', a=semi_major_axis, b=semi_major_axis,
                            lon_0=ifile.P_GAM, lat_1=ifile.P_ALP,
                            lat_2=ifile.P_BET, lat_0=ifile.YCENT)
            llcrnrlon, llcrnrlat = p(llcrnrx, llcrnry, inverse=True)
            urcrnrlon, urcrnrlat = p(urcrnrx, urcrnry, inverse=True)
            m = Basemap(projection='lcc',
                        rsphere=(semi_major_axis, semi_major_axis),
                        lon_0=ifile.P_GAM, lat_1=ifile.P_ALP,
                        lat_2=ifile.P_BET, lat_0=ifile.YCENT,
                        llcrnrlon=llcrnrlon, llcrnrlat=llcrnrlat,
                        urcrnrlat=urcrnrlat, urcrnrlon=urcrnrlon,
                        resolution=resolution, suppress_ticks=False)
        elif ifile.GDTYP == 6:
            p = getproj(ifile, withgrid=True)
            lclon, lclat = p(ifile.NCOLS / 2., 0, inverse=True)
            lllon, lllat = p(0, 0, inverse=True)
            urlon, urlat = p(ifile.NCOLS, ifile.NROWS, inverse=True)
            print(lclon, lclat)
            print(lllon, lllat)
            print(urlon, urlat)
            m = Basemap(projection='stere', lat_0=ifile.P_ALP * 90,
                        lat_ts=ifile.P_BET, lon_0=ifile.P_GAM, llcrnrlon=lllon,
                        llcrnrlat=lllat, urcrnrlon=urlon, urcrnrlat=urlat,
                        rsphere=(semi_major_axis, semi_major_axis))
        elif ifile.GDTYP == 7:
            from mpl_toolkits.basemap import pyproj
            mapstr = '+proj=merc +a=%s +b=%s +lat_ts=0 +lon_0=%s' % (
                semi_major_axis, semi_major_axis, ifile.XCENT)
            p = pyproj.Proj(mapstr)
            llcrnrlon, llcrnrlat = p(llcrnrx, llcrnry, inverse=True)
            urcrnrlon, urcrnrlat = p(urcrnrx, urcrnry, inverse=True)
            m = Basemap(projection='merc',
                        rsphere=(semi_major_axis, semi_major_axis),
                        lon_0=ifile.XCENT, lat_ts=0, llcrnrlon=llcrnrlon,
                        llcrnrlat=llcrnrlat, urcrnrlat=urcrnrlat,
                        urcrnrlon=urcrnrlon, resolution=resolution,
                        suppress_ticks=False)
        print('Found IO/API Mapping parameters')
    else:
        kwds = dict(suppress_ticks=False)
        try:
            lat, latunit = getlatbnds(ifile)
            lon, lonunit = getlonbnds(ifile)
            kwds['llcrnrlat'] = float(lat[:].min())
            kwds['urcrnrlat'] = float(lat[:].max())
            kwds['llcrnrlon'] = float(lon[:].min())
            kwds['urcrnrlon'] = float(lon[:].max())
            kwds['resolution'] = resolution
        except Exception as e:
            print(e)
            pass
        m = Basemap(**kwds)
    return m


def getinterpweights(xs, nxs, kind='linear', fill_value='extrapolate',
                     extrapolate=False):
    """
    Get weights for interpolation by matrix multiplication

    Parameters
    ----------
    xs : old input coordinates
    nxs : new output coordinates
    extrapolate : allow extrapolation beyond bounds, default False
    fill_value : set fill value (e.g, nan) to prevent extrapolation or edge
                 continuation

    Returns
    -------
    weights : numpy array shape = (old, new)

    Notes
    -----
    When extrapolate is false, the edge values are used for points beyond
    the inputs.
    Particularly useful when interpolating many values

    Example
    -------
    xs = np.arange(10, 100, 10)
    ys = xs
    nxs = np.arange(0, 100, 5)
    weights = getinterpweights(a, b)
    nys = (weights * xs[:, None]).sum(0)

    """
    from scipy.interpolate import interp1d
    # identity matrix
    ident = np.identity(xs.size)
    # weight function; use bounds outside
    weight_func = interp1d(xs, ident, axis=-1, kind='linear',
                           bounds_error=False, fill_value='extrapolate')
    # calculate weights, which can be reused
    weights = weight_func(nxs)

    # If not extrapolating, force weights to
    # no more than one at the edges
    if not extrapolate:
        weights = np.maximum(0, weights)
        weights /= weights.sum(0)
    return weights


def sigma2coeff(fromvglvls, tovglvls):
    """
    Calculate fraction of pressure from each layer in fromfile
    that is in each layer in tofile and return matrix
    """
    edges = np.interp(
        tovglvls[::-1],
        fromvglvls[::-1],
        np.arange(fromvglvls.size)[::-1])[::-1]\
        .repeat(2, 0)[1:-1].reshape(-1, 2).astype('d')
    coeff = np.zeros((fromvglvls.size - 1, tovglvls.size - 1), dtype='d')
    for li, (b, t) in enumerate(edges):
        ll = np.floor(b).astype('i')
        ul = np.ceil(t).astype('i')
        for lay in range(ll, ul):
            bf = max(b - lay, 0)
            tf = min(t - lay, 1)
            myf = min(bf, tf)
            myf = tf - bf
            coeff[lay, li] = myf
            # if li == 12:
            #    print(li, b, t, ll, ul, lay, bf, tf, min(bf, tf))
    return coeff



#!/usr/bin/env python

if __name__ == '__main__':
    from PseudoNetCDF.plotutil.pncscatter import pncscatter
    from PseudoNetCDF.pncparse import getparser, pncparse
    parser = getparser(plot_options=True, has_ofile=True)
    parser.epilog += """
    -----
pncscatter inobs inmod target [target ...]
inobs - path to obs (or other model)
inmod - path to mod (or other obs)
target - variable name
"""
    ifiles, args = pncparse(plot_options=True, has_ofile=True, parser=parser)
    pncscatter(args)



__all__ = ['vertical_diffusivity_center_time']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx vertical/diffusivity variable transformations
==================================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` variable transformations for CAMx
              vertical diffusivity files.  See PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

from PseudoNetCDF.MetaNetCDF import time_avg_new_unit
from .Memmap import vertical_diffusivity as reg_vertical_diffusivity


class vertical_diffusivity_center_time(time_avg_new_unit):
    __reader__ = reg_vertical_diffusivity



__doc__ = """
.. _Memmaps
:mod:`Memmaps` -- CAMx Memmap Interfaces
========================================

.. module:: Memmaps
   :platform: Unix, Windows
   :synopsis: Provides an easy access point for memory map based
              :ref:`PseudoNetCDF` file interfaces for all CAMx
              related files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
__all__ = ['cloud_rain',
           'height_pressure',
           'humidity',
           'ipr',
           'irr',
           'landuse',
           'one3d',
           'point_source',
           'temperature',
           'uamiv',
           'vertical_diffusivity',
           'wind',
           'lateral_boundary',
           'finst']

from .uamiv.Memmap import uamiv
from .cloud_rain.Memmap import cloud_rain
from .height_pressure.Memmap import height_pressure
from .humidity.Memmap import humidity
from .finst.Memmap import finst
from .ipr.Memmap import ipr
from .irr.Memmap import irr
from .landuse.Memmap import landuse
from .lateral_boundary.Memmap import lateral_boundary
from .point_source.Memmap import point_source
from .temperature.Memmap import temperature
from .wind.Memmap import wind
from .vertical_diffusivity.Memmap import vertical_diffusivity
from .one3d.Memmap import one3d



from __future__ import print_function
import numpy as np
from PseudoNetCDF.sci_var import PseudoNetCDFFile
from glob import glob
import sys


class flightlogs(PseudoNetCDFFile):
    def __init__(self, pathlike):
        if isinstance(pathlike, str):
            paths = glob(pathlike)
        else:
            paths = pathlike
        paths.sort()
        pfile = self
        pfile._vars = dict()

        files = [open(path) for path in paths]
        datas = [np.recfromtxt(f, names=True, case_sensitive=True)
                 for f in files]
        data = np.ma.concatenate(datas)
        desired_unit = dict(O3='ppb', GMAO_TEMP='K', PRESS='hPa', TEMP='K')
        unit_factor = {'ppt': 1e12, 'ppb': 1e9}
        pfile.createDimension('time', data.shape[0])
        for ki, key in enumerate(data.dtype.names):
            typecode = data[key].dtype.char
            if typecode not in ('c', 'S'):
                unit = desired_unit.get(key, 'ppt')
                factor = unit_factor.get(unit, 1)
                values = np.ma.masked_values(data[key], -1000) * factor
            else:
                unit = 'unknown'
                values = data[key]
            pfile.createVariable(key, typecode, dimensions=(
                'time',), units=unit, values=values)


if __name__ == '__main__':
    bfile1 = flightlogs(sys.argv[1:])
    for label, key in [('O3', 'O3[:]'), ('NO2', 'NO2[:]')]:
        bvar = eval(key, None, bfile1.variables)
        b2var = eval(key, None, bfile1.variables)
        assert((bvar == b2var).all())
        print('\n%s (BASE: %6.2f)' % (label, bvar.mean()), file=sys.stdout)
        print('\n      BASE:', sep='', file=sys.stdout)
        np.percentile(
            bvar, np.ma.arange(.1, 1., .1) * 100
        ).tofile(sys.stdout, sep=', ', format='%6.2f')
    print('', file=sys.stdout)




__all__ = ['NetCDFFile', 'NetCDFVariable']
__doc__ = """
.. _netcdf
:mod:`netcdf` -- netcdf import point
====================================

.. module:: netcdf
   :platform: Unix, Windows
   :synopsis: Povides a single import point for a package.  If
              a user has one of many netcdf interfaces, this module
              selects it and provides it.
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
from netCDF4 import Dataset as NetCDFFile
from netCDF4 import Variable as NetCDFVariable



__all__ = ['irr']
__doc__ = """
.. _Memmap
:mod:`Memmap` -- irr Memmap interface
============================================

.. module:: Memmap
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` memory map for CAMx
              irr files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct
from warnings import warn
from collections import defaultdict
from functools import partial

# Site-Packages
from numpy import zeros, memmap, dtype

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariables
from PseudoNetCDF.ArrayTransforms import ConvertCAMxTime

# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


def _isMine(path):
    testf = OpenRecordFile(path)
    if testf.record_size == 80:
        testf.next()
        if testf.record_size == 16:
            testf.next()
            if testf.record_size == 28:
                return True
    return False


class irr(PseudoNetCDFFile):
    """
    irr provides a PseudoNetCDF interface for CAMx
    irr files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> irr_path = 'camx_irr.bin'
        >>> rows,cols = 65,83
        >>> irrfile = irr(irr_path,rows,cols)
        >>> irrfile.variables.keys()
        ['TFLAG', 'RXN_01', 'RXN_02', 'RXN_03', ...]
        >>> v = irrfile.variables['RXN_01']
        >>> tflag = irrfile.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> irrfile.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "if5i"
    data_fmt = "f"

    @classmethod
    def isMine(self, path):
        return _isMine(path)

    def __init__(self, rf, multi=False):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.__rffile = OpenRecordFile(rf)
        self.__readheader()
        irr_record_type = dtype(
            dict(names=(['SPAD', 'DATE', 'TIME', 'PAGRID', 'NEST', 'I', 'J',
                         'K'] +
                        ['RXN_%02d' % i for i in range(1, self.nrxns + 1)] +
                        ['EPAD']),
                 formats=(['>i', '>i', '>f', '>i', '>i', '>i', '>i', '>i'] +
                          ['>f' for i in range(1, self.nrxns + 1)] +
                          ['>i'])))

        varkeys = [i for i in irr_record_type.names[8:-1]] + ['TFLAG']
        self.groups = defaultdict(PseudoNetCDFFile)
        padatatype = []
        pavarkeys = []
        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('VAR', len(varkeys) - 1)
        self.createDimension('DATE-TIME', 2)
        for di, domain in enumerate(self._padomains):
            dk = 'PA%02d' % (di + 1)
            prefix = dk + '_'
            pavarkeys.extend([prefix + k for k in varkeys])
            grp = self.groups[dk]
            for propk, propv in domain.items():
                setattr(grp, propk, propv)
            grp.createDimension('TSTEP', self.time_step_count)
            grp.createDimension('VAR', len(varkeys) - 1)
            grp.createDimension('DATE-TIME', 2)
            grp.createDimension('COL', domain['iend'] - domain['istart'] + 1)
            grp.createDimension('ROW', domain['jend'] - domain['jstart'] + 1)
            grp.createDimension('LAY', domain['tlay'] - domain['blay'] + 1)
            padatatype.append((dk, irr_record_type,
                               (len(grp.dimensions['ROW']),
                                len(grp.dimensions['COL']),
                                len(grp.dimensions['LAY']))))
            if len(self._padomains) == 1:
                pncols = domain['iend'] - domain['istart'] + 1
                pnrows = domain['jend'] - domain['jstart'] + 1
                pnlays = domain['tlay'] - domain['blay'] + 1
                self.createDimension('COL', pncols)
                self.createDimension('ROW', pnrows)
                self.createDimension('LAY', pnlays)
                for propk, propv in domain.items():
                    setattr(grp, propk, propv)

            varget = partial(self.__variables, dk)
            if len(self._padomains) == 1:
                self.variables = PseudoNetCDFVariables(varget, varkeys)
            else:
                grp.variables = PseudoNetCDFVariables(varget, varkeys)
        self.__memmaps = memmap(self.__rffile.infile.name, dtype(
            padatatype), 'r', self.data_start_byte)

    def __del__(self):
        try:
            del self.__memmaps
        except Exception:
            pass

    def __decorator(self, name, pncfv):
        def decor(k):
            return dict(units='ppm/hr',
                        var_desc=k.ljust(16), long_name=k.ljust(16))
        for k, v in decor(name).items():
            setattr(pncfv, k, v)
        return pncfv

    def __variables(self, pk, rxn):
        pncvar = PseudoNetCDFVariable
        if rxn == 'TFLAG':
            return ConvertCAMxTime(self.__memmaps[pk][:, 0, 0, 0]['DATE'],
                                   self.__memmaps[pk][:, 0, 0, 0]['TIME'],
                                   len(self.groups[pk].dimensions['VAR']))
        tmpvals = self.__memmaps[pk][rxn].swapaxes(1, 3).swapaxes(2, 3)
        return self.__decorator(rxn,
                                pncvar(self, rxn, 'f',
                                       ('TSTEP', 'LAY', 'ROW', 'COL'),
                                       values=tmpvals))

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        assert(self.__rffile.record_size == 80)
        self.runmessage = self.__rffile.read("80s")
        self.start_date, self.start_time, self.end_date, self.end_time = \
            self.__rffile.read("ifif")
        self.time_step = 100.
        self.time_step_count = len([i for i in self.timerange()])
        self._grids = []
        for grid in range(self.__rffile.read("i")[-1]):
            self._grids.append(
                dict(
                    zip(
                        ['orgx', 'orgy', 'ncol', 'nrow', 'xsize', 'ysize',
                         'iutm'],
                        self.__rffile.read("iiiiiii")
                    )
                )
            )

        self._padomains = []
        for padomain in range(self.__rffile.read("i")[-1]):
            self._padomains.append(
                dict(
                    zip(
                        ['grid', 'istart', 'iend', 'jstart', 'jend', 'blay',
                         'tlay'],
                        self.__rffile.read("iiiiiii")
                    )
                )
            )
        self.nrxns = self.__rffile.read('i')[-1]

        self.data_start_byte = self.__rffile.record_start
        self.record_fmt = self.id_fmt + str(self.nrxns) + self.data_fmt
        self.record_size = self.__rffile.record_size
        self.padded_size = self.record_size + 8
        domain = self._padomains[0]
        self.records_per_time = (domain['iend'] - domain['istart'] + 1) *\
                                (domain['jend'] - domain['jstart'] + 1) *\
                                (domain['tlay'] - domain['blay'] + 1)
        self.time_data_block = self.padded_size * self.records_per_time
        self.time_step = 100.

    def timerange(self):
        return timerange((self.start_date, self.start_time + self.time_step),
                         timeadd((self.end_date, self.end_time),
                         (0, self.time_step)), self.time_step)


class TestMemmap(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testIRR(self):
        warn('Test not implemented; data too big for distribution')


if __name__ == '__main__':
    unittest.main()



__all__ = ['cloud_rain_center_time',
           'cloud_rain_plus', 'cloud_rain_center_time_plus']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx cloud/rain variable transformations
==================================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` variable transformations
              for CAMx cloud/rain files. See sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""


from numpy import array

from PseudoNetCDF.MetaNetCDF import add_derived, time_avg_new_unit
from PseudoNetCDF.sci_var import PseudoNetCDFVariable
from .Memmap import cloud_rain as reg_cloud_rain


class cloud_rain_center_time(time_avg_new_unit):
    __reader__ = reg_cloud_rain


class cloud_rain_plus(add_derived):
    __childclass__ = reg_cloud_rain
    __addvars__ = ['PRECIP_RATE', 'FCLOUD']

    def __PRECIP_RATE__(self):
        if 'PRECIP' in self.variables.keys():
            val = self.variables['PRECIP']
        else:
            val = self.variables['RAIN'] + \
                self.variables['SNOW'] + self.variables['GRAUPEL']
        var = PseudoNetCDFVariable(
            self, 'PRECIP_RATE', 'f', ('TSTEP', 'LAY', 'ROW', 'COL'),
            values=(val * 10)**1.27)
        var.units = 'mm/h'
        var.long_name = 'PRECIP_RATE'.ljust(16)
        var.var_desc = 'PRECIP_RATE'.ljust(16)
        return var

    def __FCLOUD__(self):
        val = self.variables['CLOUD'] >= 5
        var = PseudoNetCDFVariable(
            self, 'FCLOUD', 'i', ('TSTEP', 'LAY', 'ROW', 'COL'),
            values=array(val, dtype='i'))
        var.units = 'None'
        var.long_name = 'FCLOUD'.ljust(16)
        var.var_desc = 'FCLOUD'.ljust(16)
        return var


class cloud_rain_center_time_plus(cloud_rain_plus):
    __childclass__ = cloud_rain_center_time



__all__ = ['height_pressure']
__doc__ = """
.. _Read
:mod:`Read` -- height_pressure Read interface
=============================================

.. module:: Read
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` random access read for CAMx
              height_pressure files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

# Distribution packages
import unittest
import struct

# Site-Packages
from numpy import zeros, array

# This Package modules
from PseudoNetCDF.camxfiles.timetuple import timediff, timeadd, timerange
from PseudoNetCDF.camxfiles.FortranFileUtil import OpenRecordFile, read_into
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables


# for use in identifying uncaught nan
listnan = struct.unpack('>f', b'\xff\xc0\x00\x00')[0]
checkarray = zeros((1,), 'f')
checkarray[0] = listnan
array_nan = checkarray[0]


class height_pressure(PseudoNetCDFFile):
    """
    height_pressure provides a PseudoNetCDF interface for CAMx
    height_pressure files.  Where possible, the inteface follows
    IOAPI conventions (see www.baronams.com).

    ex:
        >>> height_pressure_path = 'camx_height_pressure.bin'
        >>> rows,cols = 65,83
        >>> hpf = height_pressure(height_pressure_path,rows,cols)
        >>> hpf.variables.keys()
        ['TFLAG', 'HGHT', 'PRES']
        >>> v = hpf.variables['V']
        >>> tflag = hpf.variables['TFLAG']
        >>> tflag.dimensions
        ('TSTEP', 'VAR', 'DATE-TIME')
        >>> tflag[0,0,:]
        array([2005185,       0])
        >>> tflag[-1,0,:]
        array([2005185,  240000])
        >>> v.dimensions
        ('TSTEP', 'LAY', 'ROW', 'COL')
        >>> v.shape
        (25, 28, 65, 83)
        >>> hpf.dimensions
        {'TSTEP': 25, 'LAY': 28, 'ROW': 65, 'COL': 83}
    """

    id_fmt = "fi"
    data_fmt = "f"

    def __init__(self, rf, rows, cols):
        """
        Initialization included reading the header and learning
        about the format.

        see __readheader and __gettimestep() for more info
        """
        self.rffile = OpenRecordFile(rf)

        self.id_size = struct.calcsize(self.id_fmt)
        self.__readheader()
        self.__gettimestep()
        if rows is None and cols is None:
            rows = self.cell_count
            cols = 1
        elif rows is None:
            rows = self.cell_count / cols
        elif cols is None:
            cols = self.cell_count / rows
        else:
            if cols * rows != self.cell_count:
                raise ValueError(("The product of cols (%d) and rows (%d) " +
                                  "must equal cells (%d)") % (
                    cols, rows, self.cell_count))

        self.createDimension('TSTEP', self.time_step_count)
        self.createDimension('ROW', rows)
        self.createDimension('COL', cols)
        self.createDimension('LAY', self.nlayers)
        self.variables = PseudoNetCDFVariables(
            self.__var_get, ['HGHT', 'PRES'])

    def __var_get(self, key):
        def constr(hp):
            return self.getArray({'HGHT': 0, 'PRES': 1}[hp])

        def decor(hp):
            return {'HGHT': dict(units='m', var_desc='HGHT'.ljust(16),
                                 long_name='HGHT'.ljust(16)),
                    'PRES': dict(units='hPa', var_desc='PRES'.ljust(16),
                                 long_name='PRES'.ljust(16))}[hp]
        values = constr(key)
        var = self.createVariable(key, 'f', ('TSTEP', 'LAY', 'ROW', 'COL'))
        var[:] = values
        for k, v in decor(key).items():
            setattr(var, k, v)
        return var

    def __readheader(self):
        """
        __readheader reads the header section of the ipr file
        it initializes each header field (see CAMx Users Manual for a list)
        as properties of the ipr class
        """
        self.data_start_byte = 0
        self.start_time, self.start_date = self.rffile.read(self.id_fmt)
        self.record_size = self.rffile.record_size
        self.padded_size = self.record_size + 8
        self.cell_count = (
            self.record_size - self.id_size) // struct.calcsize(self.data_fmt)
        self.record_fmt = self.id_fmt + self.data_fmt * (self.cell_count)

    def __gettimestep(self):
        """
        Header information provides start and end date, but does not
        indicate the increment between.  This routine reads the first
        and second date/time and initializes variables indicating the
        timestep length and the anticipated number.
        """
        self.rffile._newrecord(
            self.padded_size * 2
        )
        d, t = self.start_date, self.start_time
        self.nlayers = 0
        while timediff((self.start_date, self.start_time), (d, t)) == 0:
            t, d = self.rffile.read(self.id_fmt)
            self.rffile.next()
            self.nlayers += 1
        self.time_step = timediff((self.start_date, self.start_time), (d, t))

        while True:
            try:
                self.seek(d, t, self.nlayers, 1, False)
                d, t = timeadd((d, t), (0, self.time_step))
            except Exception:
                break
        self.end_date, self.end_time = timeadd((d, t), (0, -self.time_step))
        self.time_step_count = int(timediff(
            (self.start_date, self.start_time),
            (self.end_date, self.end_time)) / self.time_step) + 1

    def __timerecords(self, dt):
        """
        routine returns the number of records to increment from the
        data start byte to find the first time
        """
        d, t = dt
        nsteps = int(
            timediff((self.start_date, self.start_time), (d, t)) /
            self.time_step)
        nk = self.__layerrecords(self.nlayers + 1)
        return nsteps * nk

    def __layerrecords(self, k):
        """
        routine returns the number of records to increment from the
        data start byte to find the first klayer
        """
        return (k - 1) * 2

    def __recordposition(self, date, time, k, hp):
        """
        routine uses timerecords and layerrecords multiplied plus hp
        by the fortran padded size to return the byte position of the
        specified record

        date - integer
        time - float
        k - integer
        hp - integer (0=h,1=p)
        """
        ntime = self.__timerecords((date, time))
        nk = self.__layerrecords(k)
        return (nk + ntime + hp) * self.padded_size + self.data_start_byte

    def seek(self, date=None, time=None, k=1, hp=0, chkvar=True):
        """
        Move file cursor to specified record
        """
        if date is None:
            date = self.start_date
        if time is None:
            time = self.start_time

        tchk = chkvar and \
            timediff((self.end_date, self.end_time), (date, time)) > 0 or \
            timediff((self.start_date, self.start_time), (date, time)) < 0
        if tchk:
            raise KeyError(("Vertical Diffusivity file includes (%i,%6.1f) " +
                            "thru (%i,%6.1f); you requested (%i,%6.1f)") % (
                self.start_date, self.start_time, self.end_date,
                self.end_time, date, time))
        if chkvar and k < 1 or k > self.nlayers:
            raise KeyError(
                ("Vertical Diffusivity file include layers 1 thru %i; " +
                 "you requested %i") % (self.nlayers, k))
        if chkvar and hp < 0 or hp > 1:
            raise KeyError(
                "Height pressure or indexed 0 and 1; you requested %i" % (hp))

        self.rffile._newrecord(self.__recordposition(date, time, k, hp))

    def read(self):
        """
        Call recordfile read method directly
        """
        return self.rffile.read(self.record_fmt)

    def read_into(self, dest):
        """
        put values from rffile read into dest
        dest - numpy or numeric array
        """
        return read_into(self.rffile, dest, self.id_fmt, self.data_fmt)

    def seekandreadinto(self, dest, date=None, time=None, k=1, hp=0):
        """
        see seek and read
        """
        self.seek(date, time, k, hp)
        return self.read_into(dest)

    def seekandread(self, date=None, time=None, k=1, hp=0):
        """
        see seek and read
        """
        self.seek(date, time, k, hp)
        return self.read()

    def values(self):
        for d, t, k in self.__iter__():
            yield self.seekandread(d, t, k, 0), self.seekandread(d, t, k, 1)

    def items(self):
        for d, t, k in self.__iter__():
            yield (d, t, k,
                   self.seekandread(d, t, k, 0), self.seekandread(d, t, k, 1))

    def keys(self):
        for d, t in self.timerange():
            for k in range(1, self.nlayers + 1):
                yield d, t, k

    __iter__ = keys

    def getArray(self, hp):
        a = zeros((self.time_step_count, len(self.dimensions['LAY']), len(
            self.dimensions['ROW']), len(self.dimensions['COL'])), 'f')

        for ti, (d, t) in enumerate(self.timerange()):
            for ki, k in enumerate(range(1, self.nlayers + 1)):
                self.seekandreadinto(a[ti, ki, ...], d, t, k, hp)
        return a

    def timerange(self):
        t1 = (self.start_date, self.start_time)
        tind = (2400, 24)[int(self.time_step % 2)]
        t2 = timeadd((self.end_date, self.end_time), (0, self.time_step), tind)
        return timerange(t1, t2, self.time_step, tind)


class TestRead(unittest.TestCase):
    def runTest(self):
        pass

    def setUp(self):
        pass

    def testHP(self):
        import PseudoNetCDF.testcase
        hpfile = height_pressure(
            PseudoNetCDF.testcase.camxfiles_paths['height_pressure'], 4, 5)
        checkv = array([3.38721924e+01, 3.40657959e+01, 3.41392822e+01,
                        3.42358398e+01, 3.42543945e+01, 3.38868408e+01,
                        3.40622559e+01, 3.42358398e+01, 3.44768066e+01,
                        3.46112061e+01, 3.37558594e+01, 3.39323730e+01,
                        3.42663574e+01, 3.46854248e+01, 3.48144531e+01,
                        3.39472656e+01, 3.41900635e+01, 3.46160889e+01,
                        3.48209229e+01, 3.47874756e+01, 6.78652344e+01,
                        6.82532959e+01, 6.84020996e+01, 6.85950928e+01,
                        6.86304932e+01, 6.78945312e+01, 6.82465820e+01,
                        6.85941162e+01, 6.90783691e+01, 6.93474121e+01,
                        6.76313477e+01, 6.79859619e+01, 6.86558838e+01,
                        6.94960938e+01, 6.97552490e+01, 6.80159912e+01,
                        6.85028076e+01, 6.93570557e+01, 6.97674561e+01,
                        6.97009277e+01, 1.01980713e+02, 1.02563843e+02,
                        1.02787109e+02, 1.03077759e+02, 1.03131104e+02,
                        1.02022949e+02, 1.02553101e+02, 1.03076904e+02,
                        1.03804565e+02, 1.04208984e+02, 1.01628662e+02,
                        1.02162842e+02, 1.03169922e+02, 1.04433838e+02,
                        1.04823120e+02, 1.02206909e+02, 1.02940186e+02,
                        1.04224609e+02, 1.04841797e+02, 1.04740723e+02,
                        3.38721924e+01, 3.40657959e+01, 3.41392822e+01,
                        3.42358398e+01, 3.42543945e+01, 3.38868408e+01,
                        3.40622559e+01, 3.42358398e+01, 3.44768066e+01,
                        3.46112061e+01, 3.37558594e+01, 3.39323730e+01,
                        3.42663574e+01, 3.46854248e+01, 3.48144531e+01,
                        3.39472656e+01, 3.41900635e+01, 3.46160889e+01,
                        3.48209229e+01, 3.47874756e+01, 6.78652344e+01,
                        6.82532959e+01, 6.84020996e+01, 6.85950928e+01,
                        6.86304932e+01, 6.78945312e+01, 6.82465820e+01,
                        6.85941162e+01, 6.90783691e+01, 6.93474121e+01,
                        6.76313477e+01, 6.79859619e+01, 6.86558838e+01,
                        6.94960938e+01, 6.97552490e+01, 6.80159912e+01,
                        6.85028076e+01, 6.93570557e+01, 6.97674561e+01,
                        6.97009277e+01, 1.01980713e+02, 1.02563843e+02,
                        1.02787109e+02, 1.03077759e+02, 1.03131104e+02,
                        1.02022949e+02, 1.02553101e+02, 1.03076904e+02,
                        1.03804565e+02, 1.04208984e+02, 1.01628662e+02,
                        1.02162842e+02, 1.03169922e+02, 1.04433838e+02,
                        1.04823120e+02, 1.02206909e+02, 1.02940186e+02,
                        1.04224609e+02, 1.04841797e+02, 1.04740723e+02],
                       dtype='f').reshape(2, 3, 4, 5)

        self.assertTrue((hpfile.variables['HGHT'] == checkv).all())


if __name__ == '__main__':
    unittest.main()



from ._files import PseudoNetCDFFile
from ._variables import PseudoNetCDFVariable
from PseudoNetCDF.units import convert


def PseudoNetCDFVariableConvertUnit(var, outunit):
    """
    Convert the unit of var and update the
    associated IOAPI metadata
    """
    do = PseudoNetCDFFile()
    shape = var.shape
    for i, d in enumerate(var.dimensions):
        do.createDimension(d, shape[i])
    outvar = PseudoNetCDFVariable(do, var.long_name.strip(), var.typecode(
    ), var.dimensions, values=convert(var, var.units, outunit))
    for k in var.ncattrs():
        v = getattr(var, k)
        setattr(outvar, k, v)
    outvar.units = outunit
    return outvar



import numpy as np
from PseudoNetCDF.sci_var import PseudoNetCDFFile
names = 'X,Y,CONC,ZELEV,ZHILL,ZFLAG,dum1,AVE,dum2,GRP,DATE,dum3,NETID'.split(
    ',')
units = 'm,m,micrograms/m**3,m,m,m,N/A,N/A,YYMMDDHH,N/A'.split(',')
# (3(1X,F13.5),3(1X,F8.2),2X,A6,2X,A8,2X,I8.8,2X,A8)
delimiter = [14] * 3 + [9] * 3 + [2, 6, 2, 8, 10, 2, 8]
StrLen = 10


class reader(PseudoNetCDFFile):
    @classmethod
    def isMine(cls, path, *args, **kwds):
        try:
            return open(path).read(len('* AERMOD')) == '* AERMOD'
        except Exception:
            return False

    def __init__(self, path):
        # import pdb; pdb.set_trace()
        self._data = np.recfromtxt(
            path, names=names, delimiter=delimiter, comments='*')
        self.createDimension('receptor', self._data.size)
        self.createDimension('StrLen', StrLen)
        for k, u in zip(names, units):
            if k in 'dum1 dum2 dum3'.split():
                continue
            vals = self._data[k]
            dt = vals.dtype.char
            dims = ('receptor',)
            if dt == 'S':
                vals = np.char.ljust(np.char.strip(vals), StrLen).view(
                    'S1').reshape(-1, StrLen)
                dims = ('receptor', 'StrLen')
                dt = 'S1'
            var = self.createVariable(k, dt, dims)
            var.units = u
            var[:] = vals



__all__ = ['wind_center_time_cell']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx wind variable transformations
==================================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` variable transformations
              for CAMx wind files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""
from warnings import warn
from PseudoNetCDF.sci_var import PseudoNetCDFFile, PseudoNetCDFVariables
from PseudoNetCDF.sci_var import PseudoNetCDFVariable
from PseudoNetCDF.sci_var import PseudoNetCDFVariableConvertUnit
from PseudoNetCDF.camxfiles.wind.Memmap import wind as reg_wind
from PseudoNetCDF.ArrayTransforms import CenterTime, CenterCAMxU, CenterCAMxV


class wind_center_time_cell(PseudoNetCDFFile):
    """
    CAMx Files
    """

    def __init__(self, rffile, rows, cols, outunit='m/s', endhour=True,
                 forcestaggered=False):
        self.__windfile = reg_wind(rffile, rows, cols)

        self.createDimension('TSTEP', self.__windfile.dimensions['TSTEP'] - 1)
        self.createDimension('LAY', self.__windfile.dimensions['LAY'])
        self.createDimension('ROW', self.__windfile.dimensions['ROW'])
        self.createDimension('COL', self.__windfile.dimensions['COL'])
        self.createDimension('VAR', self.__windfile.dimensions['VAR'])
        self.createDimension(
            'DATE-TIME', self.__windfile.dimensions.get('DATE-TIME', 2))
        self.__outunit = outunit
        self.__force_stagger = forcestaggered
        self.variables = PseudoNetCDFVariables(
            self.__variables, ['V', 'U', 'TFLAG'])
        self.__timeslice = {True: slice(
            1, None), False: slice(None, -1)}[endhour]

    def __variables(self, k):
        self.__add_variables()
        return self.variables[k]

    def __add_variables(self):
        v = self.createVariable(
            'TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'), keep=True)
        v[:] = self.__windfile.variables['TFLAG'][self.__timeslice]
        v.long_name = 'Time flag'
        v.units = 'DATE-TIME'
        if self.__force_stagger and self.__windfile.LSTAGGER == 0:
            warn('Cell centered values are being averaged as though ' +
                 'staggered. Could just be pre v4.3 file that was actually ' +
                 'staggered')

        for k in ['U', 'V']:
            if self.__force_stagger or self.__windfile.LSTAGGER != 0:
                if k == 'U':
                    preproc = CenterCAMxU
                elif k == 'V':
                    preproc = CenterCAMxV
            else:
                preproc = CenterTime

            var = self.__windfile.variables[k]
            v = PseudoNetCDFVariable(
                self, k, 'f', ('TSTEP', 'LAY', 'ROW', 'COL'),
                values=preproc(var))
            v.units = var.units
            v.long_name = k.ljust(16)
            v.var_desc = (k + ' at center').ljust(16)
            self.variables[k] = PseudoNetCDFVariableConvertUnit(
                v, self.__outunit)



from ..sci_var import PseudoNetCDFFile
from numpy import array, dtype, fromstring, newaxis, vectorize
from datetime import timedelta


def box_model_conc(conc_path, start_datetime):
    lines = open(conc_path, 'r').readlines()[8:]

    spc_names = [spc_name.strip() for spc_name in lines[0][:-1].split('\t')]
    data_block = ''.join(lines[1:])

    line_format = dtype(dict(names=spc_names, formats=['f'] * len(spc_names)))
    conc = fromstring(data_block, dtype='f', sep='\t').view(line_format)

    result = PseudoNetCDFFile()
    result.SDATE = int(start_datetime.strftime('%Y%j'))
    result.STIME = int(start_datetime.strftime('%H%M%S'))
    result.createDimension('TSTEP', conc.shape[0])
    result.createDimension('VAR', 1)
    result.createDimension('DATE-TIME', 2)
    for spc in spc_names:
        var = result.createVariable(spc, 'f', ('TSTEP',))
        var[:] = conc[spc]
        var.long_name = spc.ljust(16)
        var.var_desc = spc.ljust(16)
        var.units = 'ppm'

    tflag = result.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))

    t_tflag = start_datetime + \
        vectorize(lambda m: timedelta(minutes=m))(conc['Time'].astype('d'))
    tflag[:, :, 0] = vectorize(lambda d: int(
        d.strftime('%Y%j')))(t_tflag[:, newaxis])
    tflag[:, :, 1] = vectorize(lambda d: int(
        d.strftime('%H%M%S')))(t_tflag[:, newaxis])
    tflag.units = "<YYYYMMDD, HHMMSS>"
    tflag.long_name = 'TFLAG'.ljust(16)
    tflag.var_desc = 'TFLAG'.ljust(16)

    return result


def box_model_mrg(conc_path, irr_path, start_datetime):
    irr_values = open(irr_path, 'r').readlines()
    irr_values = [[float(f) for f in l.split()] for l in irr_values]
    irr_values = array(irr_values)
    lines = open(conc_path, 'r').readlines()[8:]

    spc_names = [spc_name.strip() for spc_name in lines[0][:-1].split('\t')]
    data_block = ''.join(lines[1:])

    line_format = dtype(dict(names=spc_names, formats=['f'] * len(spc_names)))

    conc = fromstring(data_block, dtype='f', sep='\t').view(line_format)
    result = PseudoNetCDFFile()
    result.SDATE = int(start_datetime.strftime('%Y%j'))
    result.STIME = int(start_datetime.strftime('%H%M%S'))
    result.createDimension('TSTEP', conc.shape[0] - 1)
    result.createDimension('PROCESS', 3)
    result.createDimension('REACTIONS', irr_values.shape[1] - 2)
    result.createDimension('SPECIES', len(spc_names) - 2)
    result.createDimension('VAR', 1)
    result.createDimension('DATE-TIME', 2)
    var = result.createVariable('IRR', 'f', ('TSTEP', 'REACTIONS'))
    var.units = 'ppm'
    var[:, :] = irr_values[:, 2:]

    var = result.createVariable('IPR', 'f', ('TSTEP', 'SPECIES', 'PROCESS'))
    var.units = 'ppm'
    var[:, :, 0] = conc.view('f').reshape(
        len(result.dimensions['TSTEP']) + 1,
        len(result.dimensions['SPECIES']) + 2)[:-1, 1:-1]

    var[:, :, 2] = conc.view('f').reshape(
        len(result.dimensions['TSTEP']) + 1,
        len(result.dimensions['SPECIES']) + 2)[1:, 1:-1]
    var[:, :, 1] = var[:, :, 2] - var[:, :, 0]

    tflag = result.createVariable('TFLAG', 'i', ('TSTEP', 'VAR', 'DATE-TIME'))

    t_tflag = start_datetime + \
        vectorize(lambda m: timedelta(minutes=m))(conc['Time'].astype('d'))
    tflag[:, :, 0] = vectorize(lambda d: int(
        d.strftime('%Y%j')))(t_tflag[1:, newaxis])
    tflag[:, :, 1] = vectorize(lambda d: int(
        d.strftime('%H%M%S')))(t_tflag[1:, newaxis])
    result.Reactions = ''.join(['IRR_%d'.ljust(16) % ri for ri in range(
        1, len(result.dimensions['REACTIONS']) + 1)])
    result.Species = ''.join([spc_name.ljust(
        16) for spc_name in spc_names if spc_name not in ['Time', 'AIR']])
    result.Process = 'INIT'.ljust(16) + 'CHEM'.ljust(16) + 'FCONC'.ljust(16)

    return result



from __future__ import print_function, unicode_literals

__all__ = ['pncrename', 'manglenames', 'removesingleton', 'getvarpnc',
           'interpvars', 'extract_from_file', 'extract_lonlat', 'mask_vals',
           'slice_dim', 'reduce_dim', 'pncfunc', 'pncbo',
           'pncbfunc', 'pncexpr', 'seqpncbo', 'mesh_dim',
           'add_attr', 'convolve_dim', 'merge', 'stack_files', 'splitdim']
import sys
from warnings import warn
import numpy as np
from collections import OrderedDict
import functools

from ._files import PseudoNetCDFFile
from ._variables import PseudoNetCDFMaskedVariable, PseudoNetCDFVariable

# Functions to be available for pncexpr
from .. import userfuncs
import datetime

if sys.version_info.major == 2:
    def bytes(x, enc):
        return str(x)
else:
    unicode = str


# TODO: replace with dynamic values
_metakeys = ['time', 'layer', 'level', 'latitude', 'longitude',
             'time_bounds', 'latitude_bounds', 'longitude_bounds',
             'ROW', 'COL', 'LAY', 'TFLAG', 'ETFLAG']


def pncrename(ifile, type_old_new):
    outf = getvarpnc(ifile, None)
    t, o, n = type_old_new.split(',')
    if t in ('d', 'dimension'):
        outf.dimensions[n] = outf.dimensions[o]
        del outf.dimensions[o]
        for k, v in outf.variables.items():
            if o in v.dimensions:
                v.dimensions = tuple(
                    [n if d == o else d for d in v.dimensions])
    if t in ('v', 'variable'):
        outf.variables[n] = outf.variables[o]
        del outf.variables[o]

    return outf


_translator = {'-': '_', '$': 'S', ' ': '_', '+': '_add_', '(': '', ')': ''}


def manglenames(f, translator=_translator):
    outf = getvarpnc(f, None)
    varkeys = list(outf.variables.items())
    for k, var in varkeys:
        nk = k
        for olds, news in _translator.items():
            nk = nk.replace(olds, news)
        if nk != k:
            outf.variables[nk] = outf.variables[k]
            del outf.variables[k]
    return outf


def removesingleton(f, rd, coordkeys=None):
    if coordkeys is None:
        coordkeys = f.getCoords()

    outf = PseudoNetCDFFile()
    for propkey in f.ncattrs():
        setattr(outf, propkey, getattr(f, propkey))
    for dk, d in f.dimensions.items():
        unlim = d.isunlimited()
        ni = len(d)
        if ni != 1 or dk != rd:
            tempd = outf.createDimension(dk, ni)
            if unlim:
                tempd.setunlimited(True)

    for vk, v in f.variables.items():
        dims = tuple([dk for dk in v.dimensions if dk in outf.dimensions])
        sdims = tuple([dk for dk in enumerate(v.dimensions)
                       if dk[1] not in outf.dimensions])[::-1]
        propd = dict([(pk, getattr(v, pk)) for pk in v.ncattrs()])
        ov = outf.createVariable(vk, v.dtype.char, dims, **propd)
        outvals = v[...]
        for di, dk in sdims:
            outvals = outvals.take(0, axis=di)

        ov[...] = outvals[...]
    return outf


def getvarpnc(f, varkeys, coordkeys=None, copy=True):
    if coordkeys is None:
        coordkeys = f.getCoords()

    coordkeys = set(coordkeys)
    if varkeys is None:
        varkeys = list(f.variables.keys())
        varkeys = [k for k in varkeys if k not in coordkeys]
    else:
        newvarkeys = list(set(varkeys).intersection(f.variables.keys()))
        newvarkeys = [varkey for varkey in varkeys if varkey in newvarkeys]
        oldvarkeys = list(set(varkeys))
        oldvarkeys.sort()
        skipping = set(oldvarkeys).difference(newvarkeys)
        if len(skipping):
            warn('Skipping %s' % ', '.join(skipping))
        varkeys = newvarkeys

    if hasattr(f, 'copy'):
        outf = f.copy(props=True, dimensions=False,
                      variables=False, data=False)
    else:
        from PseudoNetCDF.sci_var import WrapPNC
        outf = WrapPNC(f).copy(props=True, dimensions=False,
                               variables=False, data=False)

    for varkey in varkeys:
        try:
            var = eval(varkey, None, f.variables)
        except Exception:
            var = f.variables[varkey]
        for dimk, dimv in zip(var.dimensions, var.shape):
            if dimk not in outf.dimensions:
                newdimv = outf.createDimension(dimk, dimv)
                if f.dimensions[dimk].isunlimited():
                    newdimv.setunlimited(True)
                coordkeys.add(dimk)
                try:
                    tempv = f.variables[dimk]
                    if hasattr(tempv, 'bounds'):
                        coordkeys.add(tempv.bounds.strip())
                except (ValueError, KeyError, AttributeError):
                    pass
        for coordk in coordkeys:
            if coordk in f.dimensions and coordk not in outf.dimensions:
                newdimv = outf.createDimension(
                    coordk, len(f.dimensions[coordk]))
                if f.dimensions[coordk].isunlimited():
                    newdimv.setunlimited(True)

        propd = dict([(k, getattr(var, k)) for k in var.ncattrs()])
        if hasattr(var[...], 'fill_value') and 'fill_value' not in propd:
            propd['fill_value'] = var[...].fill_value

        if 'values' in propd:
            propd['pvalues'] = propd['values']
            del propd['values']
        if 'name' in propd:
            if 'standard_name' not in propd:
                propd['standard_name'] = propd['name']
            del propd['name']
        vals = var[...]
        if copy:
            vals = vals.copy()
        outf.createVariable(varkey, var.dtype.char,
                            var.dimensions, values=vals, **propd)
    for coordkey in coordkeys:
        if coordkey in f.variables.keys():
            coordvar = f.variables[coordkey]
            propd = dict([(k, getattr(coordvar, k))
                          for k in coordvar.ncattrs()])
            outf.createVariable(coordkey, coordvar.dtype.char,
                                coordvar.dimensions,
                                values=coordvar[...], **propd)
            for dk in coordvar.dimensions:
                if dk not in outf.dimensions:
                    dv = outf.createDimension(dk, len(f.dimensions[dk]))
                    dv.setunlimited(f.dimensions[dk].isunlimited())

    return outf


def interpvars(f, weights, dimension, loginterp=[]):
    """
    f - PseudoNetCDFFile
    weights - weights for new dimensions from old dimension dim(new, old)
    dimension - which dimensions will be reduced
    loginterp - iterable of keys to interp on log scale
    """
    outf = PseudoNetCDFFile()
    outf.dimensions = f.dimensions.copy()
    if hasattr(f, 'groups'):
        outf.groups = OrderedDict()
        for grpk, grpv in f.groups.items():
            outf.groups[grpk] = interpvars(grpv, weights, dimension)

    oldd = f.dimensions[dimension]
    didx, = [i for i, l in enumerate(weights.shape) if len(oldd) == l]

    newd = outf.createDimension(dimension, weights.shape[didx - 1])
    newd.setunlimited(oldd.isunlimited())
    for vark, oldvar in f.variables.items():
        if dimension in oldvar.dimensions:
            dimidx = list(oldvar.dimensions).index(dimension)
            if hasattr(oldvar, '_FillValue'):
                kwds = dict(fill_value=oldvar._FillValue,
                            missing_value=oldvar._FillValue)
            else:
                kwds = dict()
            newvar = outf.createVariable(
                vark, oldvar.dtype.char, oldvar.dimensions, **kwds)
            for ak in oldvar.ncattrs():
                setattr(newvar, ak, getattr(oldvar, ak))
            if len(weights.shape) <= len(oldvar.dimensions):
                weightslice = (None,) * (dimidx) + (Ellipsis,) + \
                    (None,) * len(oldvar.dimensions[dimidx + 1:])
            else:
                weightslice = slice(None)
            varslice = (slice(None,),) * dimidx + (None,)
            weightsv = weights[weightslice]
            oldvarv = oldvar[varslice]
            if not (weightsv.ndim - 1) == oldvar.ndim:
                warn('Wrong number of dimensions for %s' % (vark,))
            elif vark in loginterp:
                logv = np.ma.exp(
                    (weightsv * np.ma.log(oldvarv)).sum(dimidx + 1))
                newvar[:] = logv
            else:
                linv = (weightsv * oldvarv).sum(dimidx + 1)
                newvar[:] = linv
        else:
            outf.variables[vark] = oldvar
    return outf


def extract_from_file(f, lonlatfs, unique=False, gridded=None, method='nn',
                      passthrough=True):
    from ..coordutil import getlonlatcoordstr
    lonlatcoordstr = ""
    for lonlatf in lonlatfs:
        lonlatcoordstr += getlonlatcoordstr(lonlatf)
    return extract_lonlat(f, lonlatcoordstr, unique=unique, gridded=gridded,
                          method=method, passthrough=passthrough)


def extract_lonlat(f, lonlat, unique=False, gridded=None, method='nn',
                   passthrough=True):
    from PseudoNetCDF.sci_var import Pseudo2NetCDF
    try:
        from StringIO import StringIO as BytesIO
    except ImportError:
        from io import BytesIO
    import os
    outf = PseudoNetCDFFile()
    outf.dimensions = f.dimensions.copy()
    if hasattr(f, 'groups'):
        outf.groups = {}
        for grpk, grpv in f.groups.items():
            outf.groups[grpk] = extract(grpv, lonlat)

    p2p = Pseudo2NetCDF()
    p2p.verbose = 0
    p2p.addGlobalProperties(f, outf)

    if gridded is None:
        gridded = (
            ('longitude' in f.dimensions and 'latitude' in f.dimensions) or
            ('COL' in f.dimensions and 'ROW' in f.dimensions) or
            ('x' in f.dimensions and 'y' in f.dimensions)
        )
    if isinstance(lonlat, (str, )):
        lonlat = [lonlat]
    lonlatout = []
    for ll in lonlat:
        if isinstance(ll, (str, )):
            try:
                if os.path.exists(ll):
                    ll = open(ll, 'r').read().strip()
            except Exception as e:
                warn('Windows machines may have uncessary warnings; ' + str(e))

            lonlatout.append(ll)
    lonlat = ('/'.join(lonlatout))
    try:
        lons, lats = np.genfromtxt(
            BytesIO(bytes(lonlat.replace('/', '\n'), 'ASCII')),
            delimiter=',').T
    except Exception as e:
        print(str(e))
        raise e
    outf.lonlatcoords = lonlat
    if not method == 'll2ij':
        longitude = f.variables['longitude'][:]
        latitude = f.variables['latitude'][:]
        latlon1d = longitude.ndim == 1 and latitude.ndim == 1
    if method == 'll2ij':
        lonidxs, latidxs = f.ll2ij(lons, lats)

        def extractfunc(v, thiscoords):
            newslice = tuple([{'ROW': latidxs, 'COL': lonidxs,
                               'latitude': latidxs, 'longitude': lonidxs,
                               'points': latidxs,
                               'PERIM': latidxs}.get(d, slice(None))
                              for d in thiscoords])
            if newslice == ():
                return v
            else:
                return v[:][newslice]
    elif method == 'nn':
        if latlon1d and gridded:
            latitude = latitude[(slice(None), None, None)]
            longitude = longitude[(None, slice(None), None)]
        else:
            latitude = latitude[Ellipsis, None]
            longitude = longitude[Ellipsis, None]

        lonlatdims = latitude.ndim - 1
        londists = longitude - lons[(None,) * lonlatdims]
        latdists = latitude - lats[(None,) * lonlatdims]
        totaldists = ((latdists**2 + londists**2)**.5)
        if latlon1d and not gridded:
            latidxs, = lonidxs, = np.unravel_index(
                totaldists.reshape(-1, latdists.shape[-1]).argmin(0),
                totaldists.shape[:-1])
        else:
            latidxs, lonidxs = np.unravel_index(
                totaldists.reshape(-1, latdists.shape[-1]).argmin(0),
                totaldists.shape[:-1])

        def extractfunc(v, thiscoords):
            newslice = tuple([{'latitude': latidxs, 'longitude': lonidxs,
                               'points': latidxs,
                               'PERIM': latidxs}.get(d, slice(None))
                              for d in thiscoords])
            if newslice == ():
                return v
            else:
                return v[:][newslice]
    elif method == 'KDTree':
        if latlon1d and gridded:
            longitude, latitude = np.meshgrid(longitude, latitude)
        from scipy.spatial import KDTree
        tree = KDTree(np.ma.array([latitude.ravel(), longitude.ravel()]).T)
        dists, idxs = tree.query(np.ma.array([lats, lons]).T)
        if latlon1d and not gridded:
            latidxs, = lonidxs, = np.unravel_index(idxs, latitude.shape)
        else:
            latidxs, lonidxs = np.unravel_index(idxs, latitude.shape)

        def extractfunc(v, thiscoords):
            newslice = tuple([{'latitude': latidxs, 'longitude': lonidxs,
                               'points': latidxs,
                               'PERIM': latidxs}.get(d, slice(None))
                              for d in thiscoords])
            return v[newslice]
    elif method in ('linear', 'cubic'):
        from scipy.interpolate import LinearNDInterpolator
        from scipy.interpolate import CloughTocher2DInterpolator
        if method == 'cubic':
            interpclass = CloughTocher2DInterpolator
        else:
            interpclass = LinearNDInterpolator
        if latlon1d and gridded:
            longitude, latitude = np.meshgrid(longitude, latitude)
        points = np.array([longitude.ravel(), latitude.ravel()]).T

        def extractfunc(v, thiscoords):
            if 'latitude' not in thiscoords or 'longitude' not in thiscoords:
                return v
            newshape = [dl if d not in ('latitude', 'longitude') else -1
                        for di, (d, dl) in enumerate(zip(thiscoords, v.shape))]
            i1 = newshape.index(-1)
            if newshape.count(-1) > 1:
                i2 = newshape.index(-1, i1 + 1)
                assert(i1 == (i2 - 1))
                newshape.pop(i2)
            i2df = interpclass(points, np.rollaxis(
                v.reshape(*newshape), i1, 0))
            out = np.rollaxis(
                np.ma.array([i2df(lon, lat) for lat, lon in zip(lats, lons)]),
                0, len(newshape))
            return out
        latidxs = extractfunc(latitude, ('latitude', 'longitude'))
    elif method in ('cubic', 'quintic'):
        from scipy.interpolate import interp2d
        if latlon1d and gridded:
            longitude, latitude = np.meshgrid(longitude, latitude)

        def extractfunc(v, thiscoords):
            i2df = interp2d(latitude, longitude, v, kind=method)
            return np.ma.array([i2df(lat, lon)
                                for lat, lon in zip(lats, lons)])
        latidxs = extractfunc(latitude, '')
    else:
        raise ValueError('method must be: nn, KDTree')
    if unique:
        tmpx = OrderedDict()
        tmplonlat = outf.lonlatcoords.split('/')
        for lon, lat, lonlatstr in zip(lonidxs, latidxs, tmplonlat):
            if (lon, lat) not in tmpx:
                tmpx[(lon, lat)] = lonlatstr

        lonidxs, latidxs = np.array(tmpx.keys()).T
        outf.lonlatcoords_orig = outf.lonlatcoords
        outf.lonlatcoords = '/'.join([tmpx[k] for k in zip(lonidxs, latidxs)])

    for k, v in f.variables.items():
        try:
            coords = v.coordinates.split()
        except Exception:
            # special case for ioapi
            coords = [{'ROW': 'latitude', 'COL': 'longitude'}.get(
                tempdk, tempdk) for tempdk in v.dimensions]
        dims = v.dimensions
        outf.createDimension('points', len(latidxs))
        if passthrough or 'longitude' in coords or 'latitude' in coords:
            try:
                del outf.variables[k]
            except Exception:
                pass
            newdims = []
            if len(dims) != len(coords):
                thiscoords = dims
            else:
                thiscoords = coords
            for d, c in zip(dims, thiscoords):
                if (
                    d not in ('longitude', 'latitude') and
                    c not in ('longitude', 'latitude')
                ):
                    newdims.append(d)
                else:
                    if 'points' not in newdims:
                        newdims.append('points')

            newdims = tuple(newdims)
            newv = extractfunc(v, thiscoords)

            propd = dict([(ak, getattr(v, ak)) for ak in v.ncattrs()])
            nv = outf.createVariable(
                k, v.dtype.char, newdims, values=newv, **propd)
            setattr(nv, 'coordinates', getattr(
                v, 'coordinates', ' '.join(coords)))
            for di, dk in enumerate(newdims):
                if dk not in outf.dimensions:
                    outf.createDimension(dk, nv.shape[di])
    return outf


extract = extract_lonlat


def mask_vals(f, maskdef, metakeys=_metakeys):
    mtype = maskdef.split(',')[0]
    mval = ','.join(maskdef.split(',')[1:])
    if mtype == 'where':
        maskexpr = 'np.ma.masked_where(mask, var[:].view(np.ndarray))'
        # mask = eval(mval, None, f.variables)
    else:
        maskexpr = 'np.ma.masked_%s(var[:], %s)' % (mtype, mval)
    for varkey, var in f.variables.items():
        if varkey not in metakeys:
            try:
                vout = eval(maskexpr)
                f.variables[varkey] = PseudoNetCDFMaskedVariable(
                    f, varkey, var.dtype.char, var.dimensions, values=vout,
                    **dict([(pk, getattr(var, pk)) for pk in var.ncattrs()]))
            except Exception as e:
                warn('Cannot mask %s: %s' % (varkey, str(e)))
    return f


def slice_dim(f, slicedef, fuzzydim=True):
    """
    variables have dimensions (e.g., time, layer, lat, lon), which can be
    subset using:

        slice_dim(f, 'dim,start,stop,stride')

    e.g., slice_dim(f, 'layer,0,47,5') would sample every fifth layer
    starting at 0
    """
    inf = f

    historydef = "slice_dim(f, %s, fuzzydim = %s); " % (slicedef, fuzzydim)
    slicedef = slicedef.split(',')
    slicedef = [slicedef[0]] + list(map(eval, slicedef[1:]))
    if len(slicedef) == 2:
        slicedef.append(slicedef[-1] + 1)
    slicedef = (slicedef + [None, ])[:4]
    dimkey, dmin, dmax, dstride = slicedef
    if dimkey not in inf.dimensions:
        warn('%s not in file' % dimkey)
        return inf

    unlimited = inf.dimensions[dimkey].isunlimited()
    if fuzzydim:
        partial_check = [key for key in inf.dimensions if dimkey ==
                         key[:len(dimkey)] and key[len(dimkey):].isdigit()]
        for dimk in partial_check:
            inf = slice_dim(inf, '%s,%s,%s,%s' % (dimk, dmin, dmax, dstride))

    from PseudoNetCDF.sci_var import Pseudo2NetCDF
    p2p = Pseudo2NetCDF(verbose=0)
    outf = PseudoNetCDFFile()
    p2p.addDimensions(inf, outf)
    p2p.addGlobalProperties(inf, outf)

    for varkey in inf.variables.keys():
        var = inf.variables[varkey]
        if dimkey not in var.dimensions:
            p2p.addVariable(inf, outf, varkey)
        else:
            axis = list(var.dimensions).index(dimkey)
            vout = var[...].swapaxes(
                0, axis)[dmin:dmax:dstride].swapaxes(0, axis)

            newlen = vout.shape[axis]
            newdim = outf.createDimension(dimkey, newlen)
            newdim.setunlimited(unlimited)
            outf.variables[varkey] = vout

    history = getattr(outf, 'history', '')
    history += historydef
    setattr(outf, 'history', history)

    return outf


def _getfunc(a, func):
    """
    Get an approriate function that takes one optional keyword (axis)
    """
    if not hasattr(func, '__call__'):
        if hasattr(a, func):
            outfunc = getattr(a, func)
        elif isinstance(a, np.ma.MaskedArray):
            outfunc = functools.partial(getattr(np.ma, func),
                                        a, axis=None, keepdims=True)
            # outfunc = lambda axis = None, keepdims = True: getattr(
            #    np.ma, func)(a, axis=axis, keepdims=keepdims)
        elif hasattr(np, func):
            outfunc = functools.partial(getattr(np, func),
                                        a, axis=None, keepdims=True)
            # outfunc = lambda axis = None, keepdims = True: getattr(
            #     np, func)(a, axis=axis, keepdims=keepdims)
        else:
            outfunc = functools.partial(eval(func),
                                        a, axis=None, keepdims=True)
            # outfunc = lambda axis = None, keepdims = True: eval(
            #    func)(a, axis=axis, keepdims=keepdims)
    else:
        outfunc = functools.partial(np.apply_along_axis,
                                    func1d=func, arr=a,
                                    axis=None, keepdims=True)
        # outfunc = lambda axis = None, keepdims = True: np.apply_along_axis(
        #     func1d=func, axis=axis, arr=a, keepdims=keepdims)
    return outfunc


def reduce_dim(f, reducedef, fuzzydim=True, metakeys=_metakeys):
    """
    variable dimensions can be reduced using

    reduce_dim(file 'dim,function,weight')

    e.g., reduce_dim(layer,mean,weight).

    Weighting is not fully functional.
    """
    from PseudoNetCDF.sci_var import Pseudo2NetCDF
    inf = f
    metakeys = [k for k in metakeys if k in inf.variables.keys()]
    historydef = "reduce_dim(f, %s, fuzzydim = %s, metakeys = %s); " % (
        reducedef, fuzzydim, metakeys)
    import numpy as np
    if hasattr(reducedef, 'split') and hasattr(reducedef, 'count'):
        commacount = reducedef.count(',')
        reducevals = reducedef.split(',')
    else:
        commacount = len(reducedef)
        reducevals = reducedef
    if commacount == 3:
        dimkey, func, numweightkey, denweightkey = reducevals
        numweight = inf.variables[numweightkey]
        denweight = inf.variables[denweightkey]
    elif commacount == 2:
        dimkey, func, numweightkey = reducevals
        numweight = inf.variables[numweightkey]
        denweightkey = None
    elif commacount == 1:
        dimkey, func = reducevals
        numweightkey = None
        denweightkey = None
    if fuzzydim:
        partial_check = [key for key in inf.dimensions if dimkey ==
                         key[:len(dimkey)] and key[len(dimkey):].isdigit()]
        for dimk in partial_check:
            if commacount == 1:
                inf = reduce_dim(inf, '%s,%s' % (dimk, func),)
            elif commacount == 2:
                inf = reduce_dim(inf, '%s,%s,%s' % (dimk, func, numweightkey),)
            elif commacount == 3:
                inf = reduce_dim(inf, '%s,%s,%s,%s' %
                                 (dimk, func, numweightkey, denweightkey),)
    if dimkey not in inf.dimensions:
        warn('%s not in file' % dimkey)
        return inf

    p2p = Pseudo2NetCDF(verbose=0)
    outf = PseudoNetCDFFile()
    p2p.addDimensions(inf, outf)
    del outf.dimensions[dimkey]
    p2p.addGlobalProperties(inf, outf)

    # unlimited = inf.dimensions[dimkey].isunlimited()
    # outf.createDimension(dimkey, 1)
    # if unlimited:
    #    outf.dimensions[dimkey].setunlimited(True)

    for varkey in inf.variables.keys():
        var = inf.variables[varkey]
        if dimkey not in var.dimensions:
            p2p.addVariable(inf, outf, varkey)
            continue

        axis = list(var.dimensions).index(dimkey)
        # def addunitydim(var):
        #    return var[(slice(None),) * (axis + 1) + (None,)]
        vreshape = var[slice(None)]
        # vreshape = addunitydim(var)
        if varkey not in metakeys:
            if numweightkey is None:
                vout = _getfunc(vreshape, func)(axis=axis, keepdims=True)
            elif denweightkey is None:
                wvar = var * \
                    np.array(numweight, ndmin=var.ndim)[
                        (slice(None),) * axis + (slice(0, var.shape[axis]),)]
                tmpslicenum = (slice(None),) * (axis + 1) + (None,)
                vout = getattr(wvar[tmpslicenum], func)(axis=axis)
                vout.units = (vout.units.strip() + ' * ' +
                              numweight.units.strip())
                if hasattr(vout, 'base_units'):
                    vout.base_units = (vout.base_units.strip() + ' * ' +
                                       numweight.base_units.strip())
            else:
                nwvar = var * \
                    np.array(numweight, ndmin=var.ndim)[
                        (slice(None),) * axis + (slice(0, var.shape[axis]),)]
                tmpslicenum = (slice(None),) * (axis + 1) + (None,)
                tmpsliceden = ((slice(None),) * axis +
                               (slice(0, var.shape[axis]), None))
                vout = (
                    getattr(nwvar[tmpslicenum], func)(axis=axis) /
                    getattr(np.array(denweight, ndmin=var.ndim)[tmpsliceden],
                            func)(axis=axis)
                )
        else:
            if '_bounds' not in varkey and '_bnds' not in varkey:
                vout = _getfunc(vreshape, func)(axis=axis, keepdims=True)
            else:
                vout = _getfunc(vreshape, func)(axis=axis, keepdims=True)
                vmin = _getfunc(vreshape, 'min')(axis=axis, keepdims=True)
                vmax = _getfunc(vreshape, 'max')(axis=axis, keepdims=True)
                if 'lon' in varkey or 'time' in varkey:
                    try:
                        vout[..., [0, 3]] = vmin[..., [0, 3]]
                        vout[..., [1, 2]] = vmax[..., [1, 2]]
                    except Exception:
                        vout[..., [0, 1]] = vmin[0, 0], vmax[0, 1]
                elif 'lat' in varkey:
                    nmin = vout.shape[-1] // 2
                    vout[..., :nmin] = vmin[..., :nmin]
                    vout[..., nmin:] = vmax[..., nmin:]
        if dimkey not in outf.dimensions:
            outdim = outf.createDimension(dimkey, vout.shape[axis])
            outdim.setunlimited(inf.dimensions[dimkey].isunlimited())
        nvar = outf.variables[varkey] = PseudoNetCDFMaskedVariable(
            outf, varkey, var.dtype.char, var.dimensions, values=vout)
        for k in var.ncattrs():
            setattr(nvar, k, getattr(var, k))

    history = getattr(outf, 'history', '')
    history += historydef
    setattr(outf, 'history', history)
    return outf


def pncfunc(func, ifile1, coordkeys=None, verbose=0):
    """
    Perform function (func) on all variables in ifile1.  The returned file
    (rfile) contains the result

    rfile = ifile1 <op>

    func can be a function or string
    """
    from PseudoNetCDF.sci_var import Pseudo2NetCDF

    if coordkeys is None:
        coordkeys = ifile1.getCoords()
    # Copy infile1 to a temporary PseudoNetCDFFile
    p2p = Pseudo2NetCDF()
    p2p.verbose = verbose
    tmpfile = PseudoNetCDFFile()
    p2p.convert(ifile1, tmpfile)

    # For each variable, assign the new value
    # to the tmpfile variables.
    for k in tmpfile.variables.keys():
        if k in coordkeys:
            continue
        outvar = tmpfile.variables[k]
        in1var = ifile1.variables[k]
        if not hasattr(func, '__call__'):
            if hasattr(in1var, func):
                outval = getattr(in1var, func)()
            elif '.' == func[:1]:
                outval = eval('in1var[:]' + func)
        else:
            outval = func(in1var[:])
        outval = np.ma.filled(np.ma.masked_invalid(outval), -999)
        if outvar.ndim > 0:
            outvar[:] = outval
        else:
            outvar.itemset(outval)
        outvar.fill_value = -999
    return tmpfile


def pncbo(op, ifile1, ifile2, coordkeys=None, verbose=0):
    """
    Perform binary operation (op) on all variables in ifile1
    and ifile2.  The returned file (rfile) contains the result

    rfile = ifile1 <op> ifile2

    op can be any valid operator (e.g., +, -, /, *, **, &, ||)
    """
    tmpfile = ifile1.copy(props=True, dimensions=True,
                          variables=False, data=False)
    if coordkeys is None:
        coordkeys = ifile1.getCoords()

    # For each variable, assign the new value
    # to the tmpfile variables.
    for k in ifile1.variables.keys():
        in1var = ifile1.variables[k]
        if k in coordkeys:
            tmpfile.copyVariable(in1var, key=k)
        elif k not in ifile2.variables.keys():
            warn('%s not found in ifile2; copied' % k)
            tmpfile.copyVariable(in1var, key=k)
        else:
            in2var = ifile2.variables[k]
            propd = dict([(ak, getattr(in1var, ak))
                          for ak in in1var.ncattrs()])
            unit1 = getattr(in1var, 'units', 'unknown')
            unit2 = getattr(in2var, 'units', 'unknown')
            propd['units'] = '(%s) %s (%s)' % (unit1, op, unit2)
            outval = np.ma.masked_invalid(
                eval('in1var[...] %s in2var[...]' % op).view(np.ndarray))
            outvar = tmpfile.createVariable(
                k, in1var.dtype.char, in1var.dimensions, fill_value=-999,
                values=outval)
            outvar.setncatts(propd)
    return tmpfile


def pncbfunc(func, ifile1, ifile2, coordkeys=None, verbose=0):
    """
    Perform binary function (func) on all variables in ifile1
    and ifile2.  The returned file (rfile) contains the result

    rfile = ifile1 <op> ifile2

    op can be any valid operator (e.g., +, -, /, *, **, &, ||)
    """
    from PseudoNetCDF.sci_var import Pseudo2NetCDF

    if coordkeys is None:
        coordkeys = ifile1.getCoords()

    # Copy infile1 to a temporary PseudoNetCDFFile
    p2p = Pseudo2NetCDF()
    p2p.verbose = verbose
    tmpfile = PseudoNetCDFFile()
    p2p.convert(ifile1, tmpfile)

    # For each variable, assign the new value
    # to the tmpfile variables.
    for k in tmpfile.variables.keys():
        if k in coordkeys:
            continue
        outvar = tmpfile.variables[k]
        in1var = ifile1.variables[k]
        if k not in ifile2.variables.keys():
            warn('%s not found in ifile2' % k)
            continue
        in2var = ifile2.variables[k]
        outval = np.ma.filled(np.ma.masked_invalid(
            func(in1var[...], in2var[...])), -999)
        if outvar.ndim > 0:
            outvar[:] = outval
        else:
            outvar.itemset(outval)
        outvar.fill_value = -999
    return tmpfile


def _namemangler(k):
    k = k.replace('$', 'dollar')
    k = k.replace('-', 'hyphen')
    k = k.replace('(', 'lparen')
    k = k.replace(')', 'rparen')
    return k


def pncexpr(expr, ifile, verbose=0):
    """
    Evaluate an arbitrary expression in the context of ifile.variables
    and add the result to the file with appropriate units.
    """
    from PseudoNetCDF.sci_var import WrapPNC
    from symtable import symtable

    # Copy file to temporary PseudoNetCDF file
    comp = compile(expr, 'none', 'exec')

    # varkeys = [key for key in comp.co_names if key in ifile.variables]
    # getvarpnc(ifile, varkeys)
    varpnc = tmpfile = WrapPNC(ifile, None)

    # Get NetCDF variables as a dictionary with
    # names mangled to allow special characters
    # in the names
    vardict = dict([(_namemangler(k), varpnc.variables[k])
                    for k in varpnc.variables.keys()])
    # Compile the expr
    symtbl = symtable(expr, '<pncexpr>', 'exec')
    symbols = symtbl.get_symbols()
    for symbol in symbols:
        key = symbol.get_name()
        if key in vardict:
            tmpvar = vardict[key]
            break
    else:
        tmpvar = PseudoNetCDFVariable(None, 'temp', 'f', ())

    propd = dict([(k, getattr(tmpvar, k)) for k in tmpvar.ncattrs()])
    dimt = tmpvar.dimensions
    # Add all used constants as properties
    # of the output file
    vardict['ifile'] = ifile
    vardict['infile'] = ifile
    vardict['np'] = np
    vardict['datetime'] = datetime
    for fname in dir(userfuncs):
        vardict[fname] = getattr(userfuncs, fname)
    exec('from scipy.constants import *', None, vardict)
    for k in ifile.ncattrs():
        if k not in vardict:
            vardict[k] = getattr(ifile, k)
    # oldkeys = set(vardict.keys())

    # Assign expression to new variable.
    exec(comp, None, vardict)

    # newkeys = set(vardict.keys())
    # assignedkeys = [k for k in newkeys.difference(oldkeys) if k[:1] != '_']
    assignedkeys = [s.get_name() for s in symbols if s.is_assigned()]
    assignedkeys = [k for k in assignedkeys if k in vardict]
    for key in assignedkeys:
        val = vardict[key]
        # if the output variable has no dimensions, there is likely a problem
        # and the output should be defined.
        if isinstance(val, (PseudoNetCDFVariable,)) and val.dimensions != ():
            tmpfile.variables[key] = val
        else:
            tmpfile.createVariable(key, val.dtype.char,
                                   dimt, values=val, **propd)

    return tmpfile


def seqpncbo(ops, ifiles, coordkeys=None):
    for op in ops:
        ifile1, ifile2 = ifiles[:2]
        newfile = pncbo(op=op, ifile1=ifile1,
                        ifile2=ifile2, coordkeys=coordkeys)
        del ifiles[:2]
        ifiles.insert(0, newfile)
    return ifiles


def mesh_dim(f, mesh_def):
    dimkey, meshfactor, aggfunc = mesh_def.split(',')
    meshfactor = float(meshfactor)

    def spread(a, n, axis):
        return a.repeat(n, axis) * meshfactor

    try:
        aggfunc = eval(aggfunc)
    except Exception:
        aggfunc = getattr(np, aggfunc)
    if meshfactor < 1.:
        oldres = int(1. / meshfactor)
        assert(1. / meshfactor == oldres)
        newres = 1
    elif meshfactor > 1.:
        newres = int(meshfactor)
        assert(meshfactor == newres)
        oldres = 1
    from PseudoNetCDF.MetaNetCDF import newresolution
    nrf = newresolution(f, dimension=dimkey, oldres=oldres,
                        newres=newres, repeat_method=aggfunc,
                        condense_method=aggfunc)
    f.dimensions[dimkey] = nrf.dimensions[dimkey]
    for k, v in f.variables.items():
        if dimkey in v.dimensions:
            f.variables[k] = nrf.variables[k]
    return f


def add_attr(f, attr_def):
    pieces = attr_def.split(',')
    att_nm, var_nm, mode, att_typ = pieces[:4]
    att_val = ','.join(pieces[4:])
    if var_nm == 'global':
        var = f
    else:
        var = f.variables[var_nm]

    if not (att_typ == 'c' and isinstance(att_val, (str, unicode))):
        att_val = np.array(eval(att_val), dtype=att_typ)

    if mode in ('a',):
        att_val = np.append(getattr(var, att_nm, []), att_val)

    if mode in ('a', 'c', 'm', 'o'):
        setattr(var, att_nm, att_val)
    elif mode in ('d',):
        delattr(var, att_nm)
    else:
        raise KeyError('mode must be either a c m o or d')


def convolve_dim(f, convolve_def):
    convolve_parts = convolve_def.split(',')
    dimkey = convolve_parts.pop(0)
    mode = convolve_parts.pop(0)
    weights = np.array(convolve_parts, dtype='f')
    outf = PseudoNetCDFFile()
    from PseudoNetCDF.pncgen import Pseudo2NetCDF
    p2p = Pseudo2NetCDF(verbose=0)
    p2p.addGlobalProperties(f, outf)
    p2p.addDimensions(f, outf)
    dim = outf.dimensions[dimkey]
    dim = outf.createDimension(dimkey, len(
        np.convolve(weights, np.arange(len(dim)), mode=mode)))
    dim.setunlimited(f.dimensions[dimkey].isunlimited())
    for vark, var in f.variables.items():
        lconvolve = dimkey in var.dimensions
        p2p.addVariable(f, outf, vark, data=not lconvolve)
        if lconvolve:
            axisi = list(var.dimensions).index(dimkey)
            values = np.apply_along_axis(func1d=lambda x_: np.convolve(
                weights, x_, mode=mode), axis=axisi, arr=var[:])
            if isinstance(var[:], np.ma.MaskedArray):
                values = np.ma.masked_invalid(values)

            outf.variables[vark][:] = values
    return outf


def merge(fs):
    outf = getvarpnc(fs[0], None)
    for f in fs[1:]:
        for p in f.ncattrs():
            if p not in outf.ncattrs():
                setattr(outf, p, getattr(f, p))
        for d, v in f.dimensions.items():
            if d not in outf.dimensions:
                nv = outf.createDimension(d, len(v))
                nv.setunlimited(v.isunlimited())
        for k, v in f.variables.items():
            if k in outf.variables:
                if (
                    v.shape != outf.variables[k].shape or
                    not (v[...] == outf.variables[k][...]).all()
                ):
                    warn('%s already in output' % k)
            else:
                propd = dict([(p, getattr(v, p)) for p in v.ncattrs()])
                outf.createVariable(
                    k, v.dtype.char, v.dimensions, values=v, **propd)

    return outf


def stack_files(fs, stackdim, coordkeys=None):
    """
    Create files with dimensions extended by stacking.

    Currently, there is no sanity check...

    """
    f = PseudoNetCDFFile()
    tmpf = fs[0]
    if coordkeys is None:
        coordkeys = tmpf.getCoords()

    dimensions = [f_.dimensions for f_ in fs]
    shareddims = {}
    for dimk, dim in tmpf.dimensions.items():
        if dimk == stackdim:
            continue
        dimlens = map(len, [dims[dimk] for dims in dimensions])
        if all([len(dim) == i for i in dimlens]):
            shareddims[dimk] = len(dim)
    differentdims = [set(dims.keys()).difference(
        shareddims.keys()) for dims in dimensions]
    assert(all([different == set([stackdim]) for different in differentdims]))
    from PseudoNetCDF.sci_var import Pseudo2NetCDF
    p2p = Pseudo2NetCDF(verbose=0)
    p2p.addDimensions(tmpf, f)
    f.createDimension(stackdim, sum(
        [len(dims[stackdim]) for dims in dimensions]))
    p2p.addGlobalProperties(tmpf, f)
    for tmpf in fs:
        for varkey, var in tmpf.variables.items():
            if stackdim not in var.dimensions:
                if varkey in f.variables:
                    if varkey not in coordkeys:
                        warn(
                            ('Got duplicate variables for %s ' % varkey) +
                            'without stackable dimension; first value ' +
                            'retained')
                else:
                    p2p.addVariable(tmpf, f, varkey, data=True)
            else:
                if varkey not in f.variables.keys():
                    axisi = list(var.dimensions).index(stackdim)
                    values = np.ma.concatenate(
                        [f_.variables[varkey][:] for f_ in fs], axis=axisi)
                    p2p.addVariable(tmpf, f, varkey, data=False)
                    f.variables[varkey][:] = values

    return f


def splitdim(inf, olddim, newdims, newshape):
    oldsize = len(inf.dimensions[olddim])
    newsize = np.prod(newshape)
    if newsize != oldsize:
        raise ValueError('New shape, must match old dimension length: ' +
                         '%d %d %s' % (oldsize, newsize, newshape))
    if len(newdims) != len(newshape):
        raise ValueError('Shape and dimensions must match in length')
    from PseudoNetCDF.sci_var import Pseudo2NetCDF
    p2n = Pseudo2NetCDF()
    outf = PseudoNetCDFFile()
    for dk, d in inf.dimensions.items():
        if dk == olddim:
            for dk, dl in zip(newdims, newshape):
                outf.createDimension(dk, dl)
        else:
            p2n.addDimension(inf, outf, dk)

    for vk, invar in inf.variables.items():
        if olddim in invar.dimensions:
            outdims = []
            outshape = []
            for dk in invar.dimensions:
                if dk == olddim:
                    outdims.extend(newdims)
                    outshape.extend(newshape)
                else:
                    outdims.append(dk)
                    outshape.append(len(inf.dimensions[dk]))

            outvar = outf.createVariable(vk, invar.dtype.char, tuple(outdims))
            p2n.addVariableProperties(invar, outvar)
            outvar[:] = invar[:].reshape(*outshape)
        else:
            p2n.addVariable(inf, outf, vk)

    return outf



#!/usr/bin/env python
# Run this script with pnc options
from __future__ import print_function

import os

from PseudoNetCDF.pncload import PNCConsole
from PseudoNetCDF.pncparse import pncparse
import warnings
import numpy as np
import matplotlib
import matplotlib.pyplot as plt


warn = warnings.warn
Normalize = matplotlib.colors.Normalize

LogNorm = matplotlib.colors.LogNorm
SymLogNorm = matplotlib.colors.SymLogNorm
BoundaryNorm = matplotlib.colors.BoundaryNorm

LogFormatter = matplotlib.ticker.LogFormatter
ScalarFormatter = matplotlib.ticker.ScalarFormatter

fig = plt.figure()
ax = fig.add_subplot(111)


def make2ds(args):
    ifiles = args.ifiles
    cbar = None
    if len(args.figure_keywords) > 0:
        plt.setp(fig, **args.figure_keywords)
    if len(args.axes_keywords) > 0:
        plt.setp(ax, **args.axes_keywords)
    nborders = len(ax.collections)
    for fi, ifile in enumerate(ifiles):
        variables = args.variables
        if variables is None:
            variables = [key for key, var in ifile.variables.items()
                         if var.ndim == 2]
        if len(variables) == 0:
            raise ValueError('Unable to heuristically determin plottable ' +
                             'variables; use -v to specify variables for ' +
                             'plotting')

        for varkey in variables:
            var = ifile.variables[varkey]
            vals = var[:]
            if args.squeeze:
                vals = vals.squeeze()

            if args.normalize is None:
                from scipy.stats import normaltest
                vmin, vmax = vals.min(), vals.max()
                if normaltest(vals.ravel())[1] < 0.05:
                    cvals = np.ma.compressed(vals)
                    boundaries = np.percentile(cvals, np.arange(0, 110, 10))
                    warn(('Autoselect deciles colormap of %s; override ' +
                          'width --norm') % varkey)
                else:
                    boundaries = np.linspace(vmin, vmax, num=11)
                    warn(('Autoselect linear colormap of %s; ' +
                          'override width --norm') % varkey)
                ordermag = (boundaries.max() /
                            np.ma.masked_values(boundaries, 0).min())
                if (ordermag) > 10000:
                    formatter = LogFormatter(labelOnlyBase=False)
                else:
                    formatter = None
                norm = BoundaryNorm(boundaries, ncolors=256)
            else:
                norm = eval(args.normalize)
                formatter = None
            if args.colorbarformatter is not None:
                try:
                    formatter = eval(args.colorbarformatter)
                except Exception:
                    formatter = args.colorbarformatter

            vmin, vmax = vals.min(), vals.max()
            if norm.vmin is not None:
                vmin = norm.vmin
            if norm.vmax is not None:
                vmax = norm.vmax

            varunit = getattr(var, 'units', 'unknown').strip()
            vardims = [dk for dk, dv in zip(
                var.dimensions, var.shape) if dv != 1]
            print(varkey, sep='')
            del ax.collections[nborders:]
            if args.swapaxes:
                patches = ax.pcolor(vals.T, norm=norm)
                ax.set_xlabel(vardims[0])
                ax.set_ylabel(vardims[1])
            else:
                patches = ax.pcolor(vals, norm=norm)
                ax.set_xlabel(vardims[1])
                ax.set_ylabel(vardims[0])

            height = vals.shape[0]
            width = vals.shape[1]
            if width >= height:
                orientation = 'horizontal'
            else:
                orientation = 'vertical'
            if cbar is None:
                cax = None
            else:
                cax = cbar.ax
                cax.cla()

            if vals.max() > vmax and vals.min() < vmin:
                extend = 'both'
            elif vals.max() > vmax:
                extend = 'max'
            elif vals.min() < vmin:
                extend = 'min'
            else:
                extend = 'neither'
            cbar = fig.colorbar(patches, orientation=orientation,
                                cax=cax, extend=extend, format=formatter)
            del cbar.ax.texts[:]
            minmaxtxt = '; min=%.3g; max=%.3g)' % (var[:].min(), var[:].max())
            cbar.set_label(varkey + ' (' + varunit + minmaxtxt)
            # if orientation == 'vertical':
            #     cbar.ax.text(.5, 1.05, '%.3g' % var[:].max(),
            #                  horizontalalignment = 'center',
            #                  verticalalignment = 'bottom')
            #     cbar.ax.text(.5, -.06, '%.3g ' % var[:].min(),
            #                  horizontalalignment = 'center',
            #                  verticalalignment = 'top')
            # else:
            #     cbar.ax.text(1.05, .5, ' %.3g' % var[:].max(),
            #                  verticalalignment = 'center',
            #                  horizontalalignment = 'left')
            #     cbar.ax.text(-.06, .5, '%.3g ' % var[:].min(),
            #                  verticalalignment = 'center',
            #                  horizontalalignment = 'right')
            # cbar.update_ticks()
            fmt = 'png'
            outpath = args.outpath
            if len(ifiles) > 1:
                lstr = str(fi).rjust(len(str(len(ifiles))), '0')
            else:
                lstr = ''

            figpath = os.path.join(outpath + varkey + lstr + '.' + fmt)
            if args.interactive:
                csl = PNCConsole(locals=globals())
                csl.interact()

            fig.savefig(figpath)
            if args.verbose > 0:
                print('Saved fig', figpath)


if __name__ == '__main__':
    from PseudoNetCDF.pncparse import getparser
    parser = getparser(has_ofile=True, plot_options=True, interactive=True)
    # parser.add_argument('--no-squeeze', dest='squeeze', default=True,
    #                     action='store_false',
    #                     help=('Squeeze automatically removes singleton ' +
    #                           'dimensions; disabling requires user to ' +
    #                           'remove singleton dimensions with ' +
    #                           '--remove-singleton option'))
    parser.add_argument('--swapaxes', action='store_true',
                        help='Swap x-y axes')
    ifiles, args = pncparse(
        has_ofile=True, plot_options=True, interactive=True, parser=parser)
    make2ds(ifiles, args)




# THIS FILE IS GENERATED FROM PseudoNetCDF SETUP.PY
short_version = '3.2.0'
version = '3.2.0'
full_version = '3.2.0.dev0+c990327'
git_revision = 'c990327ecc68f975848e3b219602d10bf6366ac4'
release = False
if not release:
    version = full_version



__all__ = ['ncf2humidity']
__doc__ = """
.. _Write
:mod:`Write` -- CAMx humidity  writer
============================================

.. module:: Write
   :platform: Unix, Windows
   :synopsis: Provides :ref:`PseudoNetCDF` writer for CAMx
              humidity files.  See PseudoNetCDF.sci_var.PseudoNetCDFFile
              for interface details
.. moduleauthor:: Barron Henderson <barronh@unc.edu>
"""

from PseudoNetCDF.camxfiles.one3d.Write import ncf2one3d as ncf2humidity

from PseudoNetCDF._getwriter import registerwriter
registerwriter('camxfiles.humidity', ncf2humidity)
registerwriter('humidity', ncf2humidity)



#!/usr/bin/env python
from PseudoNetCDF.plotutil.vertprofile import vertprofileplot, add_vertprofile_options
from PseudoNetCDF.pncparse import pncparse, getparser

if __name__ == '__main__':
    vertparser = getparser(
        has_ofile=True, plot_options=True, interactive=False)
    add_vertprofile_options(vertparser)
    ifiles, args = pncparse(has_ofile=True, parser=vertparser)
    vertprofileplot(ifiles, args)
